from __future__ import annotations

import logging
import warnings

import pytest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient, SqlEngine
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.table_fixtures import CONFIGURED_SOURCE_TABLE_SNAPSHOT_REPOSITORY
from metricflow.test.source_schema_tools import get_populate_source_schema_shell_command
from metricflow.test.table_snapshot.table_snapshots import (
    SqlTableSnapshotRepository,
    TableSnapshotException,
)

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    argnames="table_name",
    argvalues=tuple(
        table_snapshot.table_name for table_snapshot in CONFIGURED_SOURCE_TABLE_SNAPSHOT_REPOSITORY.table_snapshots
    ),
    ids=lambda table_name: f"table_name={table_name}",
)
def test_validate_data_in_source_schema(
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    source_table_snapshot_repository: SqlTableSnapshotRepository,
    table_name: str,
    create_source_tables: None,
) -> None:
    """Verifies that the source schema contains the tables as described in the snapshot repository.

    This is useful to run when a persisted source schema is used to validate that the tables were properly created by a
    call to populate_source_schema().
    """
    if not mf_test_session_state.use_persistent_source_schema:
        pytest.skip("Skipping as this session is running without the persistent source schema flag.")

    schema_name = mf_test_session_state.mf_source_schema

    matching_table_snapshots = tuple(
        table_snapshot
        for table_snapshot in source_table_snapshot_repository.table_snapshots
        if table_snapshot.table_name == table_name
    )

    assert (
        len(matching_table_snapshots) == 1
    ), f"Did not get exactly one matching table snapshot for table name {table_name}. Got {matching_table_snapshots}"

    for table_snapshot in matching_table_snapshots:
        try:
            sql_table = SqlTable(schema_name=schema_name, table_name=table_snapshot.table_name)
            expected_table_df = table_snapshot.as_df
            actual_table_df = sql_client.query(f"SELECT * FROM {sql_table.sql}")
            assert_dataframes_equal(
                actual=actual_table_df,
                expected=expected_table_df,
                compare_names_using_lowercase=sql_client.sql_engine_type is SqlEngine.SNOWFLAKE,
            )
        except Exception as e:
            error_message = (
                f"Error verifying that a table corresponding to {table_snapshot} exists in the persistent source "
                f"schema {schema_name}. \nTry re-populating with: \n\n"
                f"{get_populate_source_schema_shell_command(sql_client.sql_engine_type)}"
            )
            # Add it to the warnings so that it stands out in a sea of test failures.
            warnings.warn(error_message)
            raise TableSnapshotException(error_message) from e



from __future__ import annotations

import logging

from metricflow.protocols.sql_client import SqlEngine
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.sql_clients.ddl_sql_client import SqlClientWithDDLMethods
from metricflow.test.table_snapshot.table_snapshots import (
    SqlTableSnapshotLoader,
    SqlTableSnapshotRepository,
)

logger = logging.getLogger(__name__)


def create_tables_listed_in_table_snapshot_repository(
    ddl_sql_client: SqlClientWithDDLMethods,
    schema_name: str,
    table_snapshot_repository: SqlTableSnapshotRepository,
) -> None:
    """Creates all tables in the table snapshot repository in the given schema."""
    snapshot_loader = SqlTableSnapshotLoader(ddl_sql_client=ddl_sql_client, schema_name=schema_name)
    for table_snapshot in table_snapshot_repository.table_snapshots:
        logger.info(f"Loading: {table_snapshot.table_name}")
        snapshot_loader.load(table_snapshot)


def get_populate_source_schema_shell_command(engine: SqlEngine) -> str:
    """Creates the environment-specific shell command for populating a source schema.

    This needs to be built in this way because different engines use different environments.
    """
    # The hatch environments are named in all lower-case, but otherwise should match the values.
    return (
        f"hatch -v run {engine.value.lower()}-env:pytest "
        f"-vv "
        f"--log-cli-level info "
        f"--use-persistent-source-schema "
        f"{__file__}::populate_source_schema"
    )


def populate_source_schema(
    mf_test_session_state: MetricFlowTestSessionState,
    ddl_sql_client: SqlClientWithDDLMethods,
    source_table_snapshot_repository: SqlTableSnapshotRepository,
) -> None:
    """Populate the source schema with the tables listed in table_snapshots.

    This can be run via pytest when this file is specified because this function was whitelisted as a "test" in
    pyproject.toml. However, because the filename does not begin with "test_", it's not normally collected and run. As
    such, all parameters to this function are defined in fixtures.
    """
    if not mf_test_session_state.use_persistent_source_schema:
        raise ValueError("This should be run with the flag to enable use of the persistent source schema")

    schema_name = mf_test_session_state.mf_source_schema

    logger.info(f"Dropping schema {schema_name}")
    ddl_sql_client.drop_schema(schema_name=schema_name, cascade=True)
    logger.info(f"Creating schema {schema_name}")
    ddl_sql_client.create_schema(schema_name=schema_name)
    create_tables_listed_in_table_snapshot_repository(
        ddl_sql_client=ddl_sql_client,
        schema_name=schema_name,
        table_snapshot_repository=source_table_snapshot_repository,
    )



from __future__ import annotations

import logging
import pathlib
from logging.handlers import TimedRotatingFileHandler
from typing import Dict, Optional

from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest

from metricflow.cli.dbt_connectors.adapter_backed_client import AdapterBackedSqlClient
from metricflow.cli.dbt_connectors.dbt_config_accessor import dbtArtifacts, dbtProjectMetadata
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.protocols.sql_client import SqlClient

logger = logging.getLogger(__name__)


class CLIContext:
    """Context for MetricFlow CLI."""

    def __init__(self) -> None:
        """Initialize the CLI context for executing commands.

        The dbt_artifacts construct must be loaded in order for logging configuration to work correctly.
        """
        self.verbose = False
        self._dbt_project_metadata: dbtProjectMetadata = dbtProjectMetadata.load_from_project_path(pathlib.Path.cwd())
        self._dbt_artifacts: Optional[dbtArtifacts] = None
        self._mf: Optional[MetricFlowEngine] = None
        self._sql_client: Optional[SqlClient] = None
        self._semantic_manifest: Optional[SemanticManifest] = None
        self._semantic_manifest_lookup: Optional[SemanticManifestLookup] = None
        # self.log_file_path invokes the dbtRunner. If this is done after the configure_logging call all of the
        # dbt CLI logging configuration could be overridden, resulting in lots of things printing to console
        self._configure_logging(log_file_path=self.log_file_path)

    def _configure_logging(self, log_file_path: pathlib.Path) -> None:
        """Initialize the logging spec for the CLI.

        This requires a fully loaded dbt project, including what amounts to a call to dbt debug.
        As a practical matter, this should not have much end user impact except in cases where they are
        using commands that do not require a working adapter AND the call to dbt debug runs slowly.
        In future we may have better API access to the log file location for the project, at which time
        we can swap this out and return to full lazy loading for any context attributes that are slow
        to initialize.
        """
        log_format = "%(asctime)s %(levelname)s %(filename)s:%(lineno)d [%(threadName)s] - %(message)s"
        logging.basicConfig(level=logging.INFO, format=log_format)

        log_file_handler = TimedRotatingFileHandler(
            filename=log_file_path,
            # Rotate every day to a new file, keep 7 days worth.
            when="D",
            interval=1,
            backupCount=7,
        )

        formatter = logging.Formatter(fmt=log_format)
        log_file_handler.setFormatter(formatter)
        log_file_handler.setLevel(logging.INFO)

        root_logger = logging.getLogger()
        # StreamHandler to the console would have been setup by logging.basicConfig
        for handler in root_logger.handlers:
            handler.setLevel(logging.CRITICAL)
        root_logger.addHandler(log_file_handler)

    @property
    def dbt_project_metadata(self) -> dbtProjectMetadata:
        """Property accessor for dbt project metadata, useful in cases where the full manifest load is not needed."""
        return self._dbt_project_metadata

    @property
    def dbt_artifacts(self) -> dbtArtifacts:
        """Property accessor for all dbt artifacts, used for powering the sql client (among other things)."""
        if self._dbt_artifacts is None:
            self._dbt_artifacts = dbtArtifacts.load_from_project_metadata(self._dbt_project_metadata)
        return self._dbt_artifacts

    @property
    def log_file_path(self) -> pathlib.Path:
        """Returns the location of the log file path for this CLI invocation."""
        # The dbt Project.log_path attribute is currently sourced from the final runtime config value accessible
        # through the CLI state flags. As such, it will deviate from the default based on the DBT_LOG_PATH environment
        # variable. Should this behavior change, we will need to update this call.
        return pathlib.Path(self._dbt_project_metadata.project.log_path, "metricflow.log")

    @property
    def sql_client(self) -> SqlClient:
        """Property accessor for the sql_client class used in the CLI."""
        if self._sql_client is None:
            self._sql_client = AdapterBackedSqlClient(self.dbt_artifacts.adapter)

        return self._sql_client

    def run_health_checks(self) -> Dict[str, Dict[str, str]]:
        """Execute the DB health checks."""
        checks_to_run = [
            ("SELECT 1", lambda: self.sql_client.execute("SELECT 1")),
        ]

        results: Dict[str, Dict[str, str]] = {}

        for step, check in checks_to_run:
            status = "SUCCESS"
            err_string = ""
            try:
                resp = check()
                logger.info(f"Health Check Item {step}: succeeded" + f" with response {str(resp)}" if resp else None)
            except Exception as e:
                status = "FAIL"
                err_string = str(e)
                logger.error(f"Health Check Item {step}: failed with error {err_string}")

            results[f"{self.sql_client.sql_engine_type} - {step}"] = {
                "status": status,
                "error_message": err_string,
            }

        return results

    @property
    def mf(self) -> MetricFlowEngine:  # noqa: D
        if self._mf is None:
            self._mf = MetricFlowEngine(
                semantic_manifest_lookup=self.semantic_manifest_lookup,
                sql_client=self.sql_client,
            )
        assert self._mf is not None
        return self._mf

    def _build_semantic_manifest_lookup(self) -> None:
        """Get the path to the models and create a corresponding SemanticManifestLookup."""
        self._semantic_manifest_lookup = SemanticManifestLookup(self.semantic_manifest)

    @property
    def semantic_manifest_lookup(self) -> SemanticManifestLookup:  # noqa: D
        if self._semantic_manifest_lookup is None:
            self._build_semantic_manifest_lookup()
        assert self._semantic_manifest_lookup is not None
        return self._semantic_manifest_lookup

    @property
    def semantic_manifest(self) -> SemanticManifest:
        """Retrieve the semantic manifest from the dbt project root."""
        return self.dbt_artifacts.semantic_manifest



from __future__ import annotations

import json
import logging
import os
from typing import Generator

import pytest
import sqlalchemy
from dbt.adapters.factory import get_adapter_by_type
from dbt.cli.main import dbtRunner

from metricflow.protocols.sql_client import SqlClient
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState, dialect_from_url
from metricflow.test.fixtures.sql_clients.adapter_backed_ddl_client import AdapterBackedDDLSqlClient
from metricflow.test.fixtures.sql_clients.common_client import SqlDialect
from metricflow.test.fixtures.sql_clients.ddl_sql_client import SqlClientWithDDLMethods

logger = logging.getLogger(__name__)


# Env vars for dbt profile configuration. We use DBT_ENV_SECRET to avoid leaking secrets in CI logs.
# Note port cannot use the secret obfuscation because it must be type-cast to int
DBT_PROFILE_PORT = "DBT_PROFILE_PORT"
DBT_ENV_SECRET_DATABASE = "DBT_ENV_SECRET_DATABASE"
DBT_ENV_SECRET_HOST = "DBT_ENV_SECRET_HOST"
DBT_ENV_SECRET_HTTP_PATH = "DBT_ENV_SECRET_HTTP_PATH"
DBT_ENV_SECRET_PASSWORD = "DBT_ENV_SECRET_PASSWORD"
DBT_ENV_SECRET_SCHEMA = "DBT_ENV_SECRET_SCHEMA"
DBT_ENV_SECRET_USER = "DBT_ENV_SECRET_USER"
DBT_ENV_SECRET_WAREHOUSE = "DBT_ENV_SECRET_WAREHOUSE"

# BigQuery is special, so it gets its own set of env vars. Keeping them split out here for consistency.
DBT_ENV_SECRET_AUTH_PROVIDER_CERT_URL = "DBT_ENV_SECRET_AUTH_PROVIDER_CERT_URL"
DBT_ENV_SECRET_AUTH_TYPE = "DBT_ENV_SECRET_AUTH_TYPE"
DBT_ENV_SECRET_AUTH_URI = "DBT_ENV_SECRET_AUTH_URI"
DBT_ENV_SECRET_CLIENT_CERT_URL = "DBT_ENV_SECRET_CLIENT_CERT_URL"
DBT_ENV_SECRET_CLIENT_EMAIL = "DBT_ENV_SECRET_CLIENT_EMAIL"
DBT_ENV_SECRET_CLIENT_ID = "DBT_ENV_SECRET_CLIENT_ID"
DBT_ENV_SECRET_PRIVATE_KEY = "DBT_ENV_SECRET_PRIVATE_KEY"
DBT_ENV_SECRET_PRIVATE_KEY_ID = "DBT_ENV_SECRET_PRIVATE_KEY_ID"
DBT_ENV_SECRET_PROJECT_ID = "DBT_ENV_SECRET_PROJECT_ID"
DBT_ENV_SECRET_TOKEN_URI = "DBT_ENV_SECRET_TOKEN_URI"


def __configure_test_env_from_url(url: str, password: str, schema: str) -> sqlalchemy.engine.URL:
    """Populates default env var mapping from a sqlalchemy URL string.

    This is used to configure the test environment from the original MF_SQL_ENGINE_URL environment variable in
    a manner compatible with the dbt profile configurations laid out for most supported engines. We return
    the parsed URL object so that individual engine configurations can override the environment variables
    as needed to match their dbt profile configuration.
    """
    parsed_url = sqlalchemy.engine.make_url(url)

    if parsed_url.drivername != "duckdb":
        assert parsed_url.host, "Engine host is not set in engine connection URL!"
        os.environ[DBT_ENV_SECRET_HOST] = parsed_url.host

    if parsed_url.username:
        os.environ[DBT_ENV_SECRET_USER] = parsed_url.username

    if parsed_url.database:
        os.environ[DBT_ENV_SECRET_DATABASE] = parsed_url.database

    if parsed_url.port:
        os.environ[DBT_PROFILE_PORT] = str(parsed_url.port)

    os.environ[DBT_ENV_SECRET_PASSWORD] = password
    os.environ[DBT_ENV_SECRET_SCHEMA] = schema

    return parsed_url


def __configure_bigquery_env_from_credential_string(password: str, schema: str) -> None:
    credential_string = password.replace("'", "")
    # `strict=False` required to work with BQ password characters.
    credentials = json.loads(credential_string, strict=False)

    assert isinstance(credentials, dict), "JSON credential string did not parse to dict type!"

    bq_env = {
        DBT_ENV_SECRET_AUTH_PROVIDER_CERT_URL: credentials.get("auth_provider_x509_cert_url"),
        DBT_ENV_SECRET_AUTH_TYPE: credentials.get("type"),
        DBT_ENV_SECRET_AUTH_URI: credentials.get("auth_uri"),
        DBT_ENV_SECRET_CLIENT_CERT_URL: credentials.get("client_x509_cert_url"),
        DBT_ENV_SECRET_CLIENT_EMAIL: credentials.get("client_email"),
        DBT_ENV_SECRET_CLIENT_ID: credentials.get("client_id"),
        DBT_ENV_SECRET_PRIVATE_KEY: credentials.get("private_key"),
        DBT_ENV_SECRET_PRIVATE_KEY_ID: credentials.get("private_key_id"),
        DBT_ENV_SECRET_PROJECT_ID: credentials.get("project_id"),
        DBT_ENV_SECRET_TOKEN_URI: credentials.get("token_uri"),
    }

    empty_keys = [k for k, v in bq_env.items() if v is None or v == ""]

    assert (
        not empty_keys
    ), f"BigQuery credentials did not contain all required values! Missing value for keys {empty_keys}."

    # Reconstruct the dict to refine the value type to str
    os.environ.update({k: v for k, v in bq_env.items() if v is not None})
    os.environ[DBT_ENV_SECRET_SCHEMA] = schema


def __configure_databricks_env_from_url(url: str, password: str, schema: str) -> None:
    """Databricks has a custom http path attribute, which we have encoded into a SqlAlchemy-like URL appendage.

    This custom parsing was ported from our original client implementation for backwards compatibility with the
    existing CI job configurations.
    """
    http_path_key = "httppath="
    url_pieces = url.split(";")
    http_paths = [url_piece for url_piece in url_pieces[1:] if url_piece.lower().startswith(http_path_key)]
    assert len(http_paths) == 1, (
        f"There should be exactly one http path in a Databricks test engine URL, but we found http paths: "
        f"{http_paths} in url {url}!"
    )
    http_path = http_paths[0].split("=")[1]
    os.environ[DBT_ENV_SECRET_HTTP_PATH] = http_path
    __configure_test_env_from_url(url_pieces[0], password=password, schema=schema)


def __initialize_dbt() -> None:
    """Invoke the dbt runner from the appropriate directory so we can fetch the relevant adapter.

    We use the debug command to initialize the profile and make it accessible. This has the nice property of
    triggering a failure with reasonable error messages in the event the dbt configs are not set up correctly.
    """
    dbt_dir = os.path.join(os.path.dirname(__file__), "dbt_projects/metricflow_testing/")
    dbtRunner().invoke(["debug"], project_dir=dbt_dir, PROFILES_DIR=dbt_dir)


def make_test_sql_client(url: str, password: str, schema: str) -> SqlClientWithDDLMethods:
    """Build SQL client based on env configs."""
    # TODO: Switch on an enum of adapter type when all engines are cut over
    dialect = dialect_from_url(url=url)

    if dialect == SqlDialect.REDSHIFT:
        __configure_test_env_from_url(url, password=password, schema=schema)
        __initialize_dbt()
        return AdapterBackedDDLSqlClient(adapter=get_adapter_by_type("redshift"))
    elif dialect == SqlDialect.SNOWFLAKE:
        parsed_url = __configure_test_env_from_url(url, password=password, schema=schema)
        assert "warehouse" in parsed_url.normalized_query, "Sql engine URL params did not include Snowflake warehouse!"
        warehouses = parsed_url.normalized_query["warehouse"]
        assert len(warehouses) == 1, f"Found more than 1 warehouse in Snowflake URL: `{warehouses}`"
        os.environ[DBT_ENV_SECRET_WAREHOUSE] = warehouses[0]
        __initialize_dbt()
        return AdapterBackedDDLSqlClient(adapter=get_adapter_by_type("snowflake"))
    elif dialect == SqlDialect.BIGQUERY:
        __configure_bigquery_env_from_credential_string(password=password, schema=schema)
        __initialize_dbt()
        return AdapterBackedDDLSqlClient(adapter=get_adapter_by_type("bigquery"))
    elif dialect == SqlDialect.POSTGRESQL:
        __configure_test_env_from_url(url, password=password, schema=schema)
        __initialize_dbt()
        return AdapterBackedDDLSqlClient(adapter=get_adapter_by_type("postgres"))
    elif dialect == SqlDialect.DUCKDB:
        __configure_test_env_from_url(url, password=password, schema=schema)
        __initialize_dbt()
        return AdapterBackedDDLSqlClient(adapter=get_adapter_by_type("duckdb"))
    elif dialect == SqlDialect.DATABRICKS:
        __configure_databricks_env_from_url(url, password=password, schema=schema)
        __initialize_dbt()
        return AdapterBackedDDLSqlClient(adapter=get_adapter_by_type("databricks"))
    else:
        raise ValueError(f"Unknown dialect: `{dialect}` in URL {url}")


@pytest.fixture(scope="session")
def ddl_sql_client(mf_test_session_state: MetricFlowTestSessionState) -> Generator[SqlClientWithDDLMethods, None, None]:
    """Provides a SqlClient with the necessary DDL and data loading methods for test configuration.

    This allows us to provide the operations necessary for executing the test suite without exposing those methods in
    MetricFlow's core SqlClient protocol.
    """
    sql_client = make_test_sql_client(
        url=mf_test_session_state.sql_engine_url,
        password=mf_test_session_state.sql_engine_password,
        schema=mf_test_session_state.mf_source_schema,
    )
    logger.info(f"Creating schema '{mf_test_session_state.mf_system_schema}'")
    sql_client.create_schema(mf_test_session_state.mf_system_schema)
    if mf_test_session_state.mf_system_schema != mf_test_session_state.mf_source_schema:
        logger.info(f"Creating schema '{mf_test_session_state.mf_source_schema}'")
        sql_client.create_schema(mf_test_session_state.mf_source_schema)

    yield sql_client

    logger.info(f"Dropping schema '{mf_test_session_state.mf_system_schema}'")
    sql_client.drop_schema(mf_test_session_state.mf_system_schema, cascade=True)
    if (
        mf_test_session_state.mf_system_schema != mf_test_session_state.mf_source_schema
        and not mf_test_session_state.use_persistent_source_schema
    ):
        logger.info(f"Dropping schema '{mf_test_session_state.mf_source_schema}'")
        sql_client.drop_schema(mf_test_session_state.mf_source_schema, cascade=True)

    sql_client.close()
    return None


@pytest.fixture(scope="session")
def sql_client(ddl_sql_client: SqlClientWithDDLMethods) -> SqlClient:
    """Provides a standard SqlClient instance for running MetricFlow tests.

    Unless the test case itself requires the DDL methods, this is the fixture we should use.
    """
    return ddl_sql_client



"""Determines whether we can compute a valid join for the given inputs.

For example:

I have a node containing a measure that's needed for a metric, and I also have a list of dimensions that I need for that
measure.

Can I join that node with other nodes using a common entity to get those dimensions? If so, what are those nodes,
what entity should I join on, and if I do that join, what dimensions can be retrieved?

Note: the term "dimension" is used below, but it actually refers to any LinkableInstance. Also, when you see the term
"start_node", think "node containing the measure". Using the term "start_node" as the scenario of joining measure nodes
to dimension nodes is an easy case to explain, but there are non-measure nodes that can be joined with dimension nodes
to realize other planned features.
"""

from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.pretty_print import pformat_big_objects

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.partitions import PartitionJoinResolver
from metricflow.dataflow.dataflow_plan import (
    BaseOutput,
    JoinDescription,
    PartitionDimensionJoinDescription,
    PartitionTimeDimensionJoinDescription,
    ValidityWindowJoinDescription,
)
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.instances import InstanceSet
from metricflow.model.semantics.semantic_model_join_evaluator import SemanticModelJoinEvaluator
from metricflow.plan_conversion.instance_converters import CreateValidityWindowJoinDescription
from metricflow.protocols.semantics import SemanticModelAccessor
from metricflow.specs.specs import (
    LinkableInstanceSpec,
    LinklessEntitySpec,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class JoinLinkableInstancesRecipe:
    """A recipe for how to join a node containing linkable instances to the "start_node".

    node_to_join contains the linkable instances that are needed - it should be filtered so that the output data set of
    that node only includes the entity instance for the join and the instances associated with
    satisfiable_linkable_specs.
    """

    node_to_join: BaseOutput
    # The entity to join "node_to_join" on.
    join_on_entity: LinklessEntitySpec
    # The linkable instances from the query that can be satisfied if we join this node. Note that this is different from
    # the linkable specs in the node that can help to satisfy the query. e.g. "user_id__country" might be one of the
    # "satisfiable_linkable_specs", but "country" is the linkable spec in the node.
    satisfiable_linkable_specs: List[LinkableInstanceSpec]

    # The partitions to join on, if there are matching partitions between the start_node and node_to_join.
    join_on_partition_dimensions: Tuple[PartitionDimensionJoinDescription, ...]
    join_on_partition_time_dimensions: Tuple[PartitionTimeDimensionJoinDescription, ...]

    validity_window: Optional[ValidityWindowJoinDescription] = None

    @property
    def join_description(self) -> JoinDescription:
        """The recipe as a join description to use in the dataflow plan node."""
        return JoinDescription(
            join_node=self.node_to_join,
            join_on_entity=self.join_on_entity,
            join_on_partition_dimensions=self.join_on_partition_dimensions,
            join_on_partition_time_dimensions=self.join_on_partition_time_dimensions,
            validity_window=self.validity_window,
        )


@dataclass(frozen=True)
class LinkableInstanceSatisfiabilityEvaluation:
    """Evaluation for a node on the ability to get linkable instances that we need to resolve a query.

    Includes whether the node contains the linkable instances that we need (local_linkable_specs), or if it can have
    them joined in (joinable_linkable_specs). Linkable instances that can't be satisfied either locally or via joins are
    listed in unjoinable_linkable_specs.
    """

    local_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    joinable_linkable_specs: Tuple[LinkableInstanceSpec, ...]
    join_recipes: Tuple[JoinLinkableInstancesRecipe, ...]
    unjoinable_linkable_specs: Tuple[LinkableInstanceSpec, ...]


class NodeEvaluatorForLinkableInstances:
    """Helps to evaluate if linkable instances can be obtained using the given node, with joins if necessary.

    For example, consider a "start_node" containing the "bookings" measure, "is_instant" dimension, and "listing_id"
    entity with nodes_available_for_joins including a node with the "listing_id" entity, and the "country"
    dimension.

    We want to know if we can get "bookings", "is_instant", "listing_id__country" using the start_node. The result
    should be that we know: "is_instant" is available locally (i.e. in the same node), and if we join another node
    containing "listing_id" and "country" by "listing_id", we can get "listing_id__country".

    Since it's used on the left side of the join, the items in the "start_node" are sometimes labeled "left".
    """

    def __init__(
        self,
        semantic_model_lookup: SemanticModelAccessor,
        nodes_available_for_joins: Sequence[BaseOutput],
        node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver,
    ) -> None:
        """Constructor.

        Args:
            semantic_model_lookup: Needed to resolve partition dimensions.
            nodes_available_for_joins: Nodes that contain linkable instances and may be joined with the "start_node"
            (e.g. the node containing a desired measure) to retrieve the needed linkable instances.
            node_data_set_resolver: Figures out what data set is output by a node.
        """
        self._semantic_model_lookup = semantic_model_lookup
        self._nodes_available_for_joins = nodes_available_for_joins
        self._node_data_set_resolver = node_data_set_resolver
        self._partition_resolver = PartitionJoinResolver(self._semantic_model_lookup)
        self._join_evaluator = SemanticModelJoinEvaluator(self._semantic_model_lookup)

    def _find_joinable_candidate_nodes_that_can_satisfy_linkable_specs(
        self,
        start_node_instance_set: InstanceSet,
        needed_linkable_specs: List[LinkableInstanceSpec],
    ) -> List[JoinLinkableInstancesRecipe]:
        """Get nodes that can be joined to get 1 or more of the "needed_linkable_specs".

        The returned list is ordered by the number of "needed_linkable_specs" that it can satisfy.
        """
        candidates_for_join: List[JoinLinkableInstancesRecipe] = []
        start_node_spec_set = start_node_instance_set.spec_set
        for right_node in self._nodes_available_for_joins:
            data_set_in_right_node: SqlDataSet = self._node_data_set_resolver.get_output_data_set(right_node)
            linkable_specs_in_right_node = data_set_in_right_node.instance_set.spec_set.linkable_specs
            entity_specs_in_right_node = data_set_in_right_node.instance_set.spec_set.entity_specs

            # For each unlinked entity in the data set, create a candidate for joining.
            # For a data set to be useful for satisfying a linkable spec, it needs to have the entity
            # and the linkable spec without the entity. This allows joining based on the entity, which will
            # then produce the linkable spec. See comments further below for more details.

            for entity_spec_in_right_node in entity_specs_in_right_node:
                # If an entity has links, what that means and whether it can be used is unclear at the moment,
                # so skip it.
                if len(entity_spec_in_right_node.entity_links) > 0:
                    continue

                entity_instance_in_right_node = None
                for instance in data_set_in_right_node.instance_set.entity_instances:
                    if instance.spec == entity_spec_in_right_node:
                        entity_instance_in_right_node = instance
                        break

                if entity_instance_in_right_node is None:
                    raise RuntimeError(f"Could not find entity instance with name ({entity_spec_in_right_node})")

                assert (
                    len(entity_instance_in_right_node.defined_from) == 1
                ), f"Did not get exactly 1 defined_from in {entity_instance_in_right_node}"

                entity_in_right_node = self._semantic_model_lookup.get_entity_in_semantic_model(
                    entity_instance_in_right_node.defined_from[0]
                )
                if entity_in_right_node is None:
                    raise RuntimeError(
                        f"Invalid SemanticModelElementReference {entity_instance_in_right_node.defined_from[0]}"
                    )

                entity_instance_in_left_node = None
                for instance in start_node_instance_set.entity_instances:
                    if instance.spec.reference == entity_spec_in_right_node.reference:
                        entity_instance_in_left_node = instance
                        break

                if entity_instance_in_left_node is None:
                    # The right node can have a superset of entities.
                    continue

                assert len(entity_instance_in_left_node.defined_from) == 1
                assert len(entity_instance_in_right_node.defined_from) == 1

                if not self._join_evaluator.is_valid_semantic_model_join(
                    left_semantic_model_reference=entity_instance_in_left_node.defined_from[0].semantic_model_reference,
                    right_semantic_model_reference=entity_instance_in_right_node.defined_from[
                        0
                    ].semantic_model_reference,
                    on_entity_reference=entity_spec_in_right_node.reference,
                ):
                    continue

                linkless_entity_spec_in_node = LinklessEntitySpec.from_element_name(
                    entity_spec_in_right_node.element_name
                )

                satisfiable_linkable_specs = []
                for needed_linkable_spec in needed_linkable_specs:
                    assert (
                        len(needed_linkable_spec.entity_links) != 0
                    ), f"Invalid needed linkable spec passed in {needed_linkable_spec}"

                    # If the entity in the data set matches the link, then it can be used for joins. For example,
                    # if the node has the entity "user_id", and dimension "country" then it can be used for
                    # satisfying "user_id__country".
                    #
                    # Multi-hop example:
                    # required_linkable_spec = "user_id__device_id__platform"
                    # entity_spec_in_data_set = "user_id"
                    #
                    # Then the data set must contain "device_id__platform", which is realized with
                    #
                    # required_linkable_spec.remove_first_entity_link()
                    #
                    # We might also need to check the entity type and see if it's the type of join we're allowing,
                    # but since we're doing all left joins now, it's been left out.

                    required_entity_matches_data_set_entity = (
                        LinklessEntitySpec.from_reference(needed_linkable_spec.entity_links[0])
                        == linkless_entity_spec_in_node
                    )
                    needed_linkable_spec_in_node = (
                        needed_linkable_spec.without_first_entity_link in linkable_specs_in_right_node
                    )
                    if required_entity_matches_data_set_entity and needed_linkable_spec_in_node:
                        satisfiable_linkable_specs.append(needed_linkable_spec)

                # If this node can satisfy some linkable specs, it could be useful to join on, so add it to the
                # candidate list.
                if len(satisfiable_linkable_specs) > 0:
                    join_on_partition_dimensions = self._partition_resolver.resolve_partition_dimension_joins(
                        start_node_spec_set=start_node_spec_set,
                        node_to_join_spec_set=data_set_in_right_node.instance_set.spec_set,
                    )
                    join_on_partition_time_dimensions = self._partition_resolver.resolve_partition_time_dimension_joins(
                        start_node_spec_set=start_node_spec_set,
                        node_to_join_spec_set=data_set_in_right_node.instance_set.spec_set,
                    )
                    validity_window_join_description = CreateValidityWindowJoinDescription(
                        self._semantic_model_lookup
                    ).transform(instance_set=data_set_in_right_node.instance_set)

                    candidates_for_join.append(
                        JoinLinkableInstancesRecipe(
                            node_to_join=right_node,
                            join_on_entity=linkless_entity_spec_in_node,
                            satisfiable_linkable_specs=satisfiable_linkable_specs,
                            join_on_partition_dimensions=join_on_partition_dimensions,
                            join_on_partition_time_dimensions=join_on_partition_time_dimensions,
                            validity_window=validity_window_join_description,
                        )
                    )

        # Return with the candidate set that can satisfy the most linkable specs at the front.
        return sorted(
            candidates_for_join,
            key=lambda x: len(x.satisfiable_linkable_specs),
            reverse=True,
        )

    @staticmethod
    def _update_candidates_that_can_satisfy_linkable_specs(
        candidates_for_join: List[JoinLinkableInstancesRecipe],
        already_satisfisfied_linkable_specs: List[LinkableInstanceSpec],
    ) -> List[JoinLinkableInstancesRecipe]:
        """Update / filter candidates_for_join based on linkable instance specs that we have already satisfied.

        Those linkable instances are no longer needed because those were satisfied previously by joining a node
        containing some needed linkable instances to the "start_node". This method will then remove the satisfied
        linkable instances from the candidate recipes, and if a candidate recipe doesn't have any linkable instances
        that can help satisfy the query, it is removed.
        """
        updated_candidate_data_sets: List[JoinLinkableInstancesRecipe] = []
        for candidate_for_join in candidates_for_join:
            updated_satisfiable_linkable_specs = list(
                set(candidate_for_join.satisfiable_linkable_specs) - set(already_satisfisfied_linkable_specs)
            )

            if len(updated_satisfiable_linkable_specs) > 0:
                updated_candidate_data_sets.append(
                    JoinLinkableInstancesRecipe(
                        node_to_join=candidate_for_join.node_to_join,
                        join_on_entity=candidate_for_join.join_on_entity,
                        satisfiable_linkable_specs=updated_satisfiable_linkable_specs,
                        join_on_partition_dimensions=candidate_for_join.join_on_partition_dimensions,
                        join_on_partition_time_dimensions=candidate_for_join.join_on_partition_time_dimensions,
                        validity_window=candidate_for_join.validity_window,
                    )
                )
        return sorted(
            updated_candidate_data_sets,
            key=lambda x: len(x.satisfiable_linkable_specs),
            reverse=True,
        )

    def evaluate_node(
        self,
        start_node: BaseOutput,
        required_linkable_specs: Sequence[LinkableInstanceSpec],
    ) -> LinkableInstanceSatisfiabilityEvaluation:
        """Evaluates if the "required_linkable_specs" can be realized by joining the "start_node" with other nodes.

        In other words, given the data set associated with "start_node":

        * Can all "required_linkable_specs" be retrieved from the start_node? (These would be considered "local").
        * If not, can they be retrieved by joining an available node though a common entity?
        * If so, return all possible ways (by joining different nodes) that can be done.
        """
        candidate_instance_set: InstanceSet = self._node_data_set_resolver.get_output_data_set(start_node).instance_set
        candidate_spec_set = candidate_instance_set.spec_set

        logger.debug(f"Candidate spec set is:\n{pformat_big_objects(candidate_spec_set)}")

        data_set_linkable_specs = candidate_spec_set.linkable_specs

        # These are linkable specs in the same data set as the measure. Those are considered "local".
        local_linkable_specs = []

        # These are linkable specs that aren't in the data set, but they might be able to be joined in.
        possibly_joinable_linkable_specs = []

        # Group required_linkable_specs into local / un-joinable / or possibly joinable.
        unjoinable_linkable_specs = []
        for required_linkable_spec in required_linkable_specs:
            is_local = required_linkable_spec in data_set_linkable_specs
            is_unjoinable = (
                len(required_linkable_spec.entity_links) == 0
                or LinklessEntitySpec.from_reference(required_linkable_spec.entity_links[0])
                not in data_set_linkable_specs
            )
            if is_local:
                local_linkable_specs.append(required_linkable_spec)
            elif is_unjoinable:
                unjoinable_linkable_specs.append(required_linkable_spec)
            else:
                possibly_joinable_linkable_specs.append(required_linkable_spec)

        candidates_for_join = self._find_joinable_candidate_nodes_that_can_satisfy_linkable_specs(
            start_node_instance_set=candidate_instance_set, needed_linkable_specs=possibly_joinable_linkable_specs
        )
        join_candidates: List[JoinLinkableInstancesRecipe] = []

        logger.info("Looping over nodes that can be joined to get the required linkable specs")

        # Using a greedy approach, try to get the "possibly_joinable_linkable_specs" by iteratively joining nodes with
        # the most matching linkable specs. We try to join nodes with the most matching specs to minimize the number of
        # joins that we have to do to. A knapsack solution is ideal, but punting on that for simplicity.
        while len(possibly_joinable_linkable_specs) > 0:
            logger.info(f"Looking for linkable specs:\n{pformat_big_objects(possibly_joinable_linkable_specs)}")

            # We've run out of candidate data sets, but there are more linkable specs that we need. That means the
            # rest of the linkable specs can't be joined in, and we're left with unjoinable specs remaining.
            if len(candidates_for_join) == 0:
                logger.info(
                    "There are no more candidate nodes that can be joined, but not all linkable specs have "
                    "been acquired."
                )
                unjoinable_linkable_specs.extend(possibly_joinable_linkable_specs)
                break

            # Join the best candidate to realize the linkable specs
            next_candidate = candidates_for_join.pop(0)
            logger.info(f"The next candidate node to be joined is:\n{pformat_big_objects(next_candidate)}")
            join_candidates.append(next_candidate)

            # Update the candidates. Since we'll be joined/ing the previously selected candidate, we no longer need
            # to get those linkable specs from the remaining candidates, and we also need to re-rank based on the new
            # requirements.
            candidates_for_join = self._update_candidates_that_can_satisfy_linkable_specs(
                candidates_for_join=candidates_for_join,
                already_satisfisfied_linkable_specs=next_candidate.satisfiable_linkable_specs,
            )

            # The once possibly joinable specs are definitely joinable and no longer need to be searched for.
            # Remove from "possibly_joinable_linkable_specs"
            possibly_joinable_linkable_specs = [
                x for x in possibly_joinable_linkable_specs if x not in next_candidate.satisfiable_linkable_specs
            ]

        logger.info("Done evaluating possible joins")
        return LinkableInstanceSatisfiabilityEvaluation(
            local_linkable_specs=tuple(local_linkable_specs),
            joinable_linkable_specs=tuple(
                itertools.chain.from_iterable([x.satisfiable_linkable_specs for x in join_candidates])
            ),
            join_recipes=tuple(join_candidates),
            unjoinable_linkable_specs=tuple(unjoinable_linkable_specs),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.protocols.query_parameter import InputOrderByParameter
from metricflow.time.date_part import DatePart


@dataclass(frozen=True)
class TimeDimensionParameter:
    """Time dimension requested in a query."""

    name: str
    grain: Optional[TimeGranularity] = None
    date_part: Optional[DatePart] = None

    def __post_init__(self) -> None:  # noqa: D
        parsed_name = StructuredLinkableSpecName.from_name(self.name)
        if parsed_name.time_granularity:
            raise ValueError("Must use object syntax for `grain` parameter if `date_part` is requested.")


@dataclass(frozen=True)
class DimensionOrEntityParameter:
    """Group by parameter requested in a query.

    Might represent an entity or a dimension.
    """

    name: str


@dataclass(frozen=True)
class MetricParameter:
    """Metric requested in a query."""

    name: str


@dataclass(frozen=True)
class OrderByParameter:
    """Order by requested in a query."""

    order_by: InputOrderByParameter
    descending: bool = False



from __future__ import annotations

import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Literal, TypedDict, Union

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from typing_extensions import NotRequired

from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.models import InferenceResult, InferenceSignalType
from metricflow.inference.renderer.base import InferenceRenderer

yaml = YAML()


class RenderedTimeColumnConfigTypeParams(TypedDict):  # noqa: D
    time_granularity: Literal["day"]


class RenderedColumnConfig(TypedDict):  # noqa: D
    name: str
    type: str
    type_params: NotRequired[RenderedTimeColumnConfigTypeParams]


class ConfigFileRenderer(InferenceRenderer):
    """Writes inference results to a set of config files."""

    UNKNOWN_FIELD_VALUE = "FIXME"

    def __init__(self, dir_path: Union[str, Path], overwrite: bool = False) -> None:
        """Initializes the renderer.

        dir_path: The path to the config directory
        overwrite: If set to False, will raise error if the directory exists
        """
        dir_path = os.path.abspath(dir_path)

        if not overwrite and os.path.exists(dir_path):
            raise ValueError("ConfigFileRender.overwrite is False but path exists.")

        if os.path.isfile(dir_path):
            raise ValueError("ConfigFileRenderer `dir_path` is a file.")

        self.dir_path = dir_path

    def _get_filename_for_table(self, table: SqlTable) -> str:
        return os.path.abspath(os.path.join(self.dir_path, f"{table.sql}.yaml"))

    def _fixme(self, comment: str) -> str:
        return f"FIXME: {comment}"

    def _render_entity_columns(self, results: List[InferenceResult]) -> List[CommentedMap]:
        type_map = {
            InferenceSignalType.ID.PRIMARY: "primary",
            InferenceSignalType.ID.FOREIGN: "foreign",
            InferenceSignalType.ID.UNIQUE: "unique",
        }

        rendered: List[CommentedMap] = [
            CommentedMap(
                {
                    "name": result.column.column_name,
                    "type": type_map.get(result.type_node, ConfigFileRenderer.UNKNOWN_FIELD_VALUE),
                }
            )
            for result in results
            if result.type_node.is_subtype_of(InferenceSignalType.ID.UNKNOWN)
        ]

        return rendered

    def _render_dimension_columns(self, results: List[InferenceResult]) -> List[CommentedMap]:
        type_map = {
            InferenceSignalType.DIMENSION.TIME: "time",
            InferenceSignalType.DIMENSION.PRIMARY_TIME: "time",
            InferenceSignalType.DIMENSION.CATEGORICAL: "categorical",
        }

        rendered: List[CommentedMap] = []
        for result in results:
            if not result.type_node.is_subtype_of(InferenceSignalType.DIMENSION.UNKNOWN):
                continue

            result_data: CommentedMap = CommentedMap(
                {
                    "name": result.column.column_name,
                    "type": type_map.get(result.type_node, ConfigFileRenderer.UNKNOWN_FIELD_VALUE),
                }
            )

            if result_data["type"] == ConfigFileRenderer.UNKNOWN_FIELD_VALUE:
                result_data.yaml_add_eol_comment(self._fixme("unknown field value"), "type")

            if result.type_node.is_subtype_of(InferenceSignalType.DIMENSION.TIME):
                type_params: CommentedMap = CommentedMap({"time_granularity": "day"})
                if result.type_node.is_subtype_of(InferenceSignalType.DIMENSION.PRIMARY_TIME):
                    type_params["is_primary"] = True
                result_data["type_params"] = type_params

            if len(result.problems) > 0:
                result_data.yaml_set_comment_before_after_key(
                    key="name",
                    before=f"{ConfigFileRenderer.UNKNOWN_FIELD_VALUE}: " + ", ".join(result.problems),
                )

            rendered.append(result_data)

        return rendered

    def _get_comments_for_unknown_columns(self, results: List[InferenceResult]) -> List[str]:
        delim = "\n  - "
        return [
            self._fixme(result.column.column_name + delim + delim.join(result.problems))
            for result in results
            if result.type_node == InferenceSignalType.UNKNOWN
        ]

    def _get_configuration_data_for_table(self, table: SqlTable, results: List[InferenceResult]) -> Dict:
        data = CommentedMap(
            {
                "semantic_model": {
                    "name": table.table_name,
                    "node_relation": {
                        "alias": table.table_name,
                        "schema_name": table.schema_name,
                        "database": table.db_name,
                        "relation_name": table.sql,
                    },
                    "entities": self._render_entity_columns(results),
                    "dimensions": self._render_dimension_columns(results),
                    "measures": [],
                }
            }
        )

        header_comments = [self._fixme("Unreviewed inferred config file")] + self._get_comments_for_unknown_columns(
            results
        )
        data.yaml_set_comment_before_after_key(
            key="semantic_model",
            before="\n".join(header_comments),
        )

        return data

    def render(self, results: List[InferenceResult]) -> None:
        """Render the inference results to files in the configured directory."""
        os.makedirs(self.dir_path, exist_ok=True)

        results_by_table: Dict[SqlTable, List[InferenceResult]] = defaultdict(list)
        for result in results:
            results_by_table[result.column.table].append(result)

        for table, results in results_by_table.items():
            table_data = self._get_configuration_data_for_table(table, results)
            table_path = self._get_filename_for_table(table)
            with open(table_path, "w") as table_file:
                yaml.dump(table_data, table_file)



from __future__ import annotations

from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataset.dataset import DataSet

# Shortcuts for referring to the metric time dimension.
MTD = DataSet.metric_time_dimension_name()
MTD_REFERENCE = DataSet.metric_time_dimension_reference()
MTD_SPEC_DAY = DataSet.metric_time_dimension_spec(TimeGranularity.DAY)
MTD_SPEC_WEEK = DataSet.metric_time_dimension_spec(TimeGranularity.WEEK)
MTD_SPEC_MONTH = DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)
MTD_SPEC_QUARTER = DataSet.metric_time_dimension_spec(TimeGranularity.QUARTER)
MTD_SPEC_YEAR = DataSet.metric_time_dimension_spec(TimeGranularity.YEAR)



from __future__ import annotations

import datetime
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Sequence, Tuple

import pandas as pd
from dbt_semantic_interfaces.implementations.elements.dimension import PydanticDimensionTypeParams
from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from dbt_semantic_interfaces.references import EntityReference, MeasureReference, MetricReference
from dbt_semantic_interfaces.type_enums import DimensionType

from metricflow.assert_one_arg import assert_exactly_one_arg_set
from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataflow.builder.node_data_set import (
    DataflowPlanNodeOutputDataSetResolver,
)
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import DataflowPlan
from metricflow.dataflow.optimizer.source_scan.source_scan_optimizer import (
    SourceScanOptimizer,
)
from metricflow.dataflow.sql_table import SqlTable
from metricflow.dataset.convert_semantic_model import SemanticModelToDataSetConverter
from metricflow.dataset.dataset import DataSet
from metricflow.dataset.semantic_model_adapter import SemanticModelDataSet
from metricflow.engine.models import Dimension, Entity, Measure, Metric
from metricflow.engine.time_source import ServerTimeSource
from metricflow.errors.errors import ExecutionException
from metricflow.execution.execution_plan import ExecutionPlan, SqlQuery
from metricflow.execution.execution_plan_to_text import execution_plan_to_text
from metricflow.execution.executor import SequentialPlanExecutor
from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.formatting import indent_log_line
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.model.semantics.linkable_element_properties import (
    LinkableElementProperties,
)
from metricflow.model.semantics.linkable_spec_resolver import LinkableDimension
from metricflow.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_execution import (
    DataflowToExecutionPlanConverter,
)
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.protocols.query_parameter import GroupByParameter, MetricQueryParameter, OrderByQueryParameter
from metricflow.protocols.sql_client import SqlClient
from metricflow.query.query_exceptions import InvalidQueryException
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.random_id import random_id
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.specs import InstanceSpecSet, MetricFlowQuerySpec
from metricflow.sql.optimizer.optimization_levels import SqlQueryOptimizationLevel
from metricflow.telemetry.models import TelemetryLevel
from metricflow.telemetry.reporter import TelemetryReporter, log_call
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_source import TimeSource

logger = logging.getLogger(__name__)
_telemetry_reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
_telemetry_reporter.add_python_log_handler()
_telemetry_reporter.add_rudderstack_handler()


@dataclass(frozen=True)
class MetricFlowRequestId:
    """Uniquely identifies a request to the MF engine."""

    mf_rid: str


class MetricFlowQueryType(Enum):
    """An enum to designate what type of MetricFlow query it is."""

    METRIC = "metric"
    DIMENSION_VALUES = "dimension_values"


@dataclass(frozen=True)
class MetricFlowQueryRequest:
    """Encapsulates the parameters for a metric query.

    metric_names: Names of the metrics to query.
    metrics: Metric objects to query.
    group_by_names: Names of the dimensions and entities to query.
    group_by: Dimension or entity objects to query.
    limit: Limit the result to this many rows.
    time_constraint_start: Get data for the start of this time range.
    time_constraint_end: Get data for the end of this time range.
    where_constraint: A SQL string using group by names that can be used like a where clause on the output data.
    order_by_names: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds".
    order_by: metric, dimension, or entity objects to order by.
    output_table: If specified, output the result data to this table instead of a result dataframe.
    sql_optimization_level: The level of optimization for the generated SQL.
    query_type: Type of MetricFlow query.
    """

    request_id: MetricFlowRequestId
    metric_names: Optional[Sequence[str]] = None
    metrics: Optional[Sequence[MetricQueryParameter]] = None
    group_by_names: Optional[Sequence[str]] = None
    group_by: Optional[Tuple[GroupByParameter, ...]] = None
    limit: Optional[int] = None
    time_constraint_start: Optional[datetime.datetime] = None
    time_constraint_end: Optional[datetime.datetime] = None
    where_constraint: Optional[str] = None
    order_by_names: Optional[Sequence[str]] = None
    order_by: Optional[Sequence[OrderByQueryParameter]] = None
    output_table: Optional[str] = None
    sql_optimization_level: SqlQueryOptimizationLevel = SqlQueryOptimizationLevel.O4
    query_type: MetricFlowQueryType = MetricFlowQueryType.METRIC

    @staticmethod
    def create_with_random_request_id(  # noqa: D
        metric_names: Optional[Sequence[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        group_by_names: Optional[Sequence[str]] = None,
        group_by: Optional[Tuple[GroupByParameter, ...]] = None,
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraint: Optional[str] = None,
        order_by_names: Optional[Sequence[str]] = None,
        order_by: Optional[Sequence[OrderByQueryParameter]] = None,
        output_table: Optional[str] = None,
        sql_optimization_level: SqlQueryOptimizationLevel = SqlQueryOptimizationLevel.O4,
        query_type: MetricFlowQueryType = MetricFlowQueryType.METRIC,
    ) -> MetricFlowQueryRequest:
        assert_exactly_one_arg_set(metric_names=metric_names, metrics=metrics)
        assert not (
            group_by_names and group_by
        ), "Both group_by_names and group_by were set, but if a group by is specified you should only use one of these!"
        assert not (
            order_by_names and order_by
        ), "Both order_by_names and order_by were set, but if an order by is specified you should only use one of these!"
        return MetricFlowQueryRequest(
            request_id=MetricFlowRequestId(mf_rid=f"{random_id()}"),
            metric_names=metric_names,
            metrics=metrics,
            group_by_names=group_by_names,
            group_by=group_by,
            limit=limit,
            time_constraint_start=time_constraint_start,
            time_constraint_end=time_constraint_end,
            where_constraint=where_constraint,
            order_by_names=order_by_names,
            order_by=order_by,
            output_table=output_table,
            sql_optimization_level=sql_optimization_level,
            query_type=query_type,
        )


@dataclass(frozen=True)
class MetricFlowQueryResult:  # noqa: D
    """The result of a query and context on how it was generated."""

    query_spec: MetricFlowQuerySpec
    dataflow_plan: DataflowPlan
    sql: str
    result_df: Optional[pd.DataFrame] = None
    result_table: Optional[SqlTable] = None


@dataclass(frozen=True)
class MetricFlowExplainResult:
    """Returns plans for resolving a query."""

    query_spec: MetricFlowQuerySpec
    dataflow_plan: DataflowPlan
    execution_plan: ExecutionPlan
    output_table: Optional[SqlTable] = None

    @property
    def rendered_sql(self) -> SqlQuery:
        """Return the SQL query that would be run for the given query."""
        if len(self.execution_plan.tasks) != 1:
            raise NotImplementedError(
                f"Multiple tasks in the execution plan not yet supported. Got tasks: {self.execution_plan.tasks}"
            )

        sql_query = self.execution_plan.tasks[0].sql_query
        if not sql_query:
            raise NotImplementedError(
                f"Execution plan tasks without a SQL query not yet supported. Got tasks: {self.execution_plan.tasks}"
            )

        return sql_query

    @property
    def rendered_sql_without_descriptions(self) -> SqlQuery:
        """Return the SQL query without the inline descriptions."""
        sql_query = self.rendered_sql
        return SqlQuery(
            sql_query="\n".join(
                filter(
                    lambda line: not line.strip().startswith("--"),
                    sql_query.sql_query.split("\n"),
                )
            ),
            bind_parameters=sql_query.bind_parameters,
        )


class AbstractMetricFlowEngine(ABC):
    """Query interface for clients."""

    @abstractmethod
    def query(
        self,
        mf_request: MetricFlowQueryRequest,
    ) -> MetricFlowQueryResult:
        """Query for metrics."""
        pass

    @abstractmethod
    def explain(
        self,
        mf_request: MetricFlowQueryRequest,
    ) -> MetricFlowExplainResult:
        """Similar to query - returns the query that would have been executed."""
        pass

    @abstractmethod
    def simple_dimensions_for_metrics(
        self, metric_names: List[str], without_any_property: Sequence[LinkableElementProperties]
    ) -> List[Dimension]:
        """Retrieves a list of all common dimensions for metric_names.

        "simple" dimensions are the ones that people expect from a UI perspective. For example, if "ds" is a time
        dimension at a day granularity, this would not list "ds__week".

        Args:
            metric_names: Names of metrics to get common dimensions from.
            without_any_property: Return dimensions not matching these properties.

        Returns:
            A list of Dimension objects containing metadata.
        """
        pass

    @abstractmethod
    def entities_for_metrics(self, metric_names: List[str]) -> List[Entity]:
        """Retrieves a list of all entities for metric_names.

        Args:
            metric_names: Names of metrics to get common entities from.

        Returns:
            A list of Entity objects containing metadata.
        """
        pass

    @abstractmethod
    def list_metrics(self) -> List[Metric]:
        """Retrieves a list of metric names.

        Returns:
            A list of Metric objects containing metadata.
        """
        pass

    @abstractmethod
    def get_dimension_values(
        self,
        metric_names: List[str],
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:
        """Retrieves a list of dimension values given a [metric_name, get_group_by_values].

        Args:
            metric_name: Names of metrics that contain the group_by.
            get_group_by_values: Name of group_by to get values from.
            time_constraint_start: Get data for the start of this time range.
            time_constraint_end: Get data for the end of this time range.

        Returns:
            A list of dimension values as string.
        """
        pass

    @abstractmethod
    def explain_get_dimension_values(  # noqa: D
        self,
        metric_names: Optional[List[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        get_group_by_values: Optional[str] = None,
        group_by: Optional[GroupByParameter] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> MetricFlowExplainResult:
        """Returns the SQL query for get_dimension_values.

        Args:
            metric_name: Names of metrics that contain the group_by.
            get_group_by_values: Name of group_by to get values from.
            time_constraint_start: Get data for the start of this time range.
            time_constraint_end: Get data for the end of this time range.

        Returns:
            An object with the rendered SQL and generated plans.
        """
        pass


class MetricFlowEngine(AbstractMetricFlowEngine):
    """Main entry point for queries.

    Attributes on this class should be treated as in use by our APIs.
    TODO: provide a more stable API layer instead of assuming this class is stable.
    """

    def __init__(
        self,
        semantic_manifest_lookup: SemanticManifestLookup,
        sql_client: SqlClient,
        time_source: TimeSource = ServerTimeSource(),
        column_association_resolver: Optional[ColumnAssociationResolver] = None,
    ) -> None:
        """Initializer for MetricFlowEngine.

        For direct calls to construct MetricFlowEngine, do not pass the following parameters,
        - time_source
        - column_association_resolver
        - time_spine_source
        These parameters are mainly there to be overridden during tests.
        """
        self._semantic_manifest_lookup = semantic_manifest_lookup
        self._sql_client = sql_client
        self._column_association_resolver = column_association_resolver or (
            DunderColumnAssociationResolver(semantic_manifest_lookup)
        )
        self._time_source = time_source
        self._time_spine_source = semantic_manifest_lookup.time_spine_source

        self._source_data_sets: List[SemanticModelDataSet] = []
        converter = SemanticModelToDataSetConverter(column_association_resolver=self._column_association_resolver)
        for semantic_model in sorted(
            self._semantic_manifest_lookup.semantic_manifest.semantic_models, key=lambda model: model.name
        ):
            data_set = converter.create_sql_source_data_set(semantic_model)
            self._source_data_sets.append(data_set)
            logger.info(f"Created source dataset from semantic model '{semantic_model.name}'")

        source_node_builder = SourceNodeBuilder(self._semantic_manifest_lookup)
        source_nodes = source_node_builder.create_from_data_sets(self._source_data_sets)
        read_nodes = source_node_builder.create_read_nodes_from_data_sets(self._source_data_sets)

        node_output_resolver = DataflowPlanNodeOutputDataSetResolver(
            column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup),
            semantic_manifest_lookup=semantic_manifest_lookup,
        )

        self._dataflow_plan_builder = DataflowPlanBuilder(
            source_nodes=source_nodes,
            read_nodes=read_nodes,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )
        self._to_sql_query_plan_converter = DataflowToSqlQueryPlanConverter(
            column_association_resolver=self._column_association_resolver,
            semantic_manifest_lookup=self._semantic_manifest_lookup,
        )
        self._to_execution_plan_converter = DataflowToExecutionPlanConverter(
            sql_plan_converter=self._to_sql_query_plan_converter,
            sql_plan_renderer=self._sql_client.sql_query_plan_renderer,
            sql_client=sql_client,
        )
        self._executor = SequentialPlanExecutor()

        self._query_parser = MetricFlowQueryParser(
            column_association_resolver=self._column_association_resolver,
            model=self._semantic_manifest_lookup,
            read_nodes=read_nodes,
            node_output_resolver=node_output_resolver,
        )

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def query(self, mf_request: MetricFlowQueryRequest) -> MetricFlowQueryResult:  # noqa: D
        logger.info(f"Starting query request:\n" f"{indent_log_line(pformat_big_objects(mf_request))}")
        explain_result = self._create_execution_plan(mf_request)
        execution_plan = explain_result.execution_plan

        if len(execution_plan.tasks) != 1:
            raise NotImplementedError("Multiple tasks not yet supported.")

        task = execution_plan.tasks[0]

        logger.info(f"Sequentially running tasks in:\n" f"{execution_plan_to_text(execution_plan)}")
        execution_results = self._executor.execute_plan(execution_plan)
        logger.info("Finished running tasks in execution plan")

        if execution_results.contains_task_errors:
            raise ExecutionException(f"Got errors while executing tasks:\n{execution_results.get_result(task.task_id)}")

        task_execution_result = execution_results.get_result(task.task_id)

        assert task_execution_result.sql, "Task execution should have returned SQL that was run"

        logger.info(f"Finished query request: {mf_request.request_id}")
        return MetricFlowQueryResult(
            query_spec=explain_result.query_spec,
            dataflow_plan=explain_result.dataflow_plan,
            sql=task_execution_result.sql,
            result_df=task_execution_result.df,
            result_table=explain_result.output_table,
        )

    @property
    def all_time_constraint(self) -> TimeRangeConstraint:
        """TimeRangeConstraint representing the min & max dates supported."""
        return TimeRangeConstraint.all_time()

    def _create_execution_plan(self, mf_query_request: MetricFlowQueryRequest) -> MetricFlowExplainResult:
        query_spec = self._query_parser.parse_and_validate_query(
            metric_names=mf_query_request.metric_names,
            metrics=mf_query_request.metrics,
            group_by_names=mf_query_request.group_by_names,
            group_by=mf_query_request.group_by,
            limit=mf_query_request.limit,
            time_constraint_start=mf_query_request.time_constraint_start,
            time_constraint_end=mf_query_request.time_constraint_end,
            where_constraint_str=mf_query_request.where_constraint,
            order_by_names=mf_query_request.order_by_names,
            order_by=mf_query_request.order_by,
        )
        logger.info(f"Query spec is:\n{pformat_big_objects(query_spec)}")

        if self._semantic_manifest_lookup.metric_lookup.contains_cumulative_or_time_offset_metric(
            tuple(m.as_reference for m in query_spec.metric_specs)
        ):
            if self._time_spine_source.time_column_granularity != TimeGranularity.DAY:
                raise RuntimeError(
                    f"A time spine source with a granularity {self._time_spine_source.time_column_granularity} is not "
                    f"yet supported."
                )
            logger.warning(
                f"Query spec requires a time spine dataset conforming to the following spec: {self._time_spine_source}. "
            )
            time_constraint_updated = False
            if not mf_query_request.time_constraint_start:
                time_constraint_start = self._time_source.get_time() - datetime.timedelta(days=365)
                logger.warning(
                    "A start time has not be supplied while querying for cumulative metrics. To avoid an excessive "
                    f"number of rows, the start time will be changed to {time_constraint_start.isoformat()}"
                )
                time_constraint_updated = True
            if not mf_query_request.time_constraint_end:
                time_constraint_end = self._time_source.get_time()
                logger.warning(
                    "A end time has not be supplied while querying for cumulative metrics. To avoid an excessive "
                    f"number of rows, the end time will be changed to {time_constraint_end.isoformat()}"
                )
                time_constraint_updated = True
            if time_constraint_updated:
                query_spec = self._query_parser.parse_and_validate_query(
                    metric_names=mf_query_request.metric_names,
                    group_by_names=mf_query_request.group_by_names,
                    group_by=mf_query_request.group_by,
                    limit=mf_query_request.limit,
                    time_constraint_start=mf_query_request.time_constraint_start,
                    time_constraint_end=mf_query_request.time_constraint_end,
                    where_constraint_str=mf_query_request.where_constraint,
                    order_by_names=mf_query_request.order_by_names,
                    order_by=mf_query_request.order_by,
                )
                logger.warning(f"Query spec updated to:\n{pformat_big_objects(query_spec)}")

        output_table: Optional[SqlTable] = None
        if mf_query_request.output_table is not None:
            output_table = SqlTable.from_string(mf_query_request.output_table)

        output_selection_specs: Optional[InstanceSpecSet] = None
        if mf_query_request.query_type == MetricFlowQueryType.DIMENSION_VALUES:
            # Filter result by dimension columns if it's a dimension values query
            if len(query_spec.entity_specs) > 0:
                raise InvalidQueryException("Querying dimension values for entities is not allowed.")
            output_selection_specs = InstanceSpecSet(
                dimension_specs=query_spec.dimension_specs,
                time_dimension_specs=query_spec.time_dimension_specs,
            )

        if query_spec.metric_specs:
            dataflow_plan = self._dataflow_plan_builder.build_plan(
                query_spec=query_spec,
                output_sql_table=output_table,
                output_selection_specs=output_selection_specs,
                optimizers=(SourceScanOptimizer(),),
            )
        else:
            dataflow_plan = self._dataflow_plan_builder.build_plan_for_distinct_values(query_spec=query_spec)

        if len(dataflow_plan.sink_output_nodes) > 1:
            raise NotImplementedError(
                f"Multiple output nodes in the dataflow plan not yet supported. "
                f"Got tasks: {dataflow_plan.sink_output_nodes}"
            )

        execution_plan = self._to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

        return MetricFlowExplainResult(
            query_spec=query_spec,
            dataflow_plan=dataflow_plan,
            execution_plan=execution_plan,
            output_table=output_table,
        )

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def explain(self, mf_request: MetricFlowQueryRequest) -> MetricFlowExplainResult:  # noqa: D
        return self._create_execution_plan(mf_request)

    def get_measures_for_metrics(self, metric_names: List[str]) -> List[Measure]:  # noqa: D
        metrics = self._semantic_manifest_lookup.metric_lookup.get_metrics(
            metric_references=[MetricReference(element_name=metric_name) for metric_name in metric_names]
        )
        semantic_model_lookup = self._semantic_manifest_lookup.semantic_model_lookup

        measures = set()
        for metric in metrics:
            for input_measure in metric.input_measures:
                measure_reference = MeasureReference(element_name=input_measure.name)
                # populate new obj
                measure = semantic_model_lookup.get_measure(measure_reference=measure_reference)
                measures.add(
                    Measure(
                        name=measure.name,
                        agg=measure.agg,
                        agg_time_dimension=semantic_model_lookup.get_agg_time_dimension_for_measure(
                            measure_reference=measure_reference
                        ).element_name,
                        description=measure.description,
                        expr=measure.expr,
                        agg_params=measure.agg_params,
                    )
                )
        return list(measures)

    def simple_dimensions_for_metrics(  # noqa: D
        self,
        metric_names: List[str],
        without_any_property: Sequence[LinkableElementProperties] = (
            LinkableElementProperties.ENTITY,
            LinkableElementProperties.DERIVED_TIME_GRANULARITY,
            LinkableElementProperties.LOCAL_LINKED,
        ),
    ) -> List[Dimension]:
        path_key_to_linkable_dimensions = (
            self._semantic_manifest_lookup.metric_lookup.linkable_set_for_metrics(
                metric_references=[MetricReference(element_name=mname) for mname in metric_names],
                without_any_property=frozenset(without_any_property),
            )
        ).path_key_to_linkable_dimensions

        dimensions: List[Dimension] = []
        linkable_dimensions_tuple: Tuple[LinkableDimension, ...]
        for (
            path_key,
            linkable_dimensions_tuple,
        ) in path_key_to_linkable_dimensions.items():
            for linkable_dimension in linkable_dimensions_tuple:
                semantic_model = self._semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
                    linkable_dimension.semantic_model_origin
                )
                assert semantic_model

                if LinkableElementProperties.METRIC_TIME in linkable_dimension.properties:
                    metric_time_name = DataSet.metric_time_dimension_name()
                    assert linkable_dimension.element_name == metric_time_name, (
                        f"{linkable_dimension} has the {LinkableElementProperties.METRIC_TIME}, but the name does not"
                        f"match."
                    )

                    dimensions.append(
                        Dimension(
                            name=metric_time_name,
                            qualified_name=StructuredLinkableSpecName(
                                element_name=metric_time_name,
                                entity_link_names=tuple(
                                    entity_reference.element_name
                                    for entity_reference in linkable_dimension.entity_links
                                ),
                                time_granularity=linkable_dimension.time_granularity,
                            ).qualified_name,
                            description="Event time for metrics.",
                            metadata=None,
                            type_params=PydanticDimensionTypeParams(
                                time_granularity=linkable_dimension.time_granularity,
                                validity_params=None,
                            ),
                            is_partition=False,
                            type=DimensionType.TIME,
                        )
                    )
                else:
                    dimensions.append(
                        Dimension.from_pydantic(
                            pydantic_dimension=SemanticModelLookup.get_dimension_from_semantic_model(
                                semantic_model=semantic_model,
                                dimension_reference=linkable_dimension.reference,
                            ),
                            path_key=path_key,
                        )
                    )
        return sorted(dimensions, key=lambda dimension: dimension.qualified_name)

    def entities_for_metrics(self, metric_names: List[str]) -> List[Entity]:  # noqa: D
        path_key_to_linkable_entities = (
            self._semantic_manifest_lookup.metric_lookup.linkable_set_for_metrics(
                metric_references=[MetricReference(element_name=mname) for mname in metric_names],
                with_any_property=frozenset(
                    {
                        LinkableElementProperties.ENTITY,
                    }
                ),
            )
        ).path_key_to_linkable_entities

        entities: List[Entity] = []
        for (
            path_key,
            linkable_entity_tuple,
        ) in path_key_to_linkable_entities.items():
            for linkable_entity in linkable_entity_tuple:
                semantic_model = self._semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
                    linkable_entity.semantic_model_origin
                )
                assert semantic_model
                entities.append(
                    Entity.from_pydantic(
                        pydantic_entity=SemanticModelLookup.get_entity_from_semantic_model(
                            semantic_model=semantic_model,
                            entity_reference=EntityReference(element_name=linkable_entity.element_name),
                        )
                    )
                )
        return entities

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def list_metrics(self) -> List[Metric]:  # noqa: D
        metric_references = self._semantic_manifest_lookup.metric_lookup.metric_references
        metrics = self._semantic_manifest_lookup.metric_lookup.get_metrics(metric_references)
        return [
            Metric.from_pydantic(
                pydantic_metric=metric,
                dimensions=self.simple_dimensions_for_metrics([metric.name]),
            )
            for metric in metrics
        ]

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def get_dimension_values(  # noqa: D
        self,
        metric_names: List[str],
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:
        # Run query
        query_result = self.query(
            MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=metric_names,
                group_by_names=[get_group_by_values],
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                query_type=MetricFlowQueryType.DIMENSION_VALUES,
            )
        )
        result_dataframe = query_result.result_df
        if result_dataframe is None:
            return []

        # Snowflake likes upper-casing things in result output, so we lower-case all names
        # before operating on the dataframe.
        metric_names = [metric_name.lower() for metric_name in metric_names]
        result_dataframe.columns = result_dataframe.columns.str.lower()

        # Get dimension values regardless of input name -> output dimension mapping. This is necessary befcause
        # granularity adjustments on time dimensions produce different output names for dimension values.
        # Note: this only works as long as we have exactly one column of group by values
        # and no other extraneous output columns
        dim_vals = result_dataframe[result_dataframe.columns[~result_dataframe.columns.isin(metric_names)]].iloc[:, 0]

        return sorted([str(val) for val in dim_vals])

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def explain_get_dimension_values(  # noqa: D
        self,
        metric_names: Optional[List[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        get_group_by_values: Optional[str] = None,
        group_by: Optional[GroupByParameter] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> MetricFlowExplainResult:
        assert not (
            get_group_by_values and group_by
        ), "Both get_group_by_values and group_by were set, but if a group by is specified you should only use one of these!"
        return self._create_execution_plan(
            MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=metric_names,
                metrics=metrics,
                group_by_names=(get_group_by_values,) if get_group_by_values else None,
                group_by=(group_by,) if group_by else None,
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                query_type=MetricFlowQueryType.DIMENSION_VALUES,
            )
        )



from __future__ import annotations

from enum import Enum
from typing import FrozenSet


class LinkableElementProperties(Enum):
    """The properties associated with a valid linkable element.

    Local means an element that is defined within the same semantic model as the measure. This definition is used
    throughout the related classes.
    """

    # A local element as per above definition.
    LOCAL = "local"
    # A local dimension that is prefixed with a local primary entity.
    LOCAL_LINKED = "local_linked"
    # An element that was joined to the measure semantic model by an entity.
    JOINED = "joined"
    # An element that was joined to the measure semantic model by joining multiple semantic models.
    MULTI_HOP = "multi_hop"
    # A time dimension that is a version of a time dimension in a semantic model, but at a different granularity.
    DERIVED_TIME_GRANULARITY = "derived_time_granularity"
    # Refers to an entity, not a dimension.
    ENTITY = "entity"
    # See metric_time in DataSet
    METRIC_TIME = "metric_time"

    @staticmethod
    def all_properties() -> FrozenSet[LinkableElementProperties]:  # noqa: D
        return frozenset(
            {
                LinkableElementProperties.LOCAL,
                LinkableElementProperties.LOCAL_LINKED,
                LinkableElementProperties.JOINED,
                LinkableElementProperties.MULTI_HOP,
                LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                LinkableElementProperties.METRIC_TIME,
            }
        )



from __future__ import annotations

import pytest

from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter


@pytest.fixture(scope="session")
def dataflow_to_sql_converter(  # noqa: D
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> DataflowToSqlQueryPlanConverter:
    return DataflowToSqlQueryPlanConverter(
        column_association_resolver=DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
    )



from __future__ import annotations

import logging
import time
from typing import Optional

import pandas as pd

from metricflow.cli.dbt_connectors.adapter_backed_client import AdapterBackedSqlClient
from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlEngine

logger = logging.getLogger(__name__)


class AdapterBackedDDLSqlClient(AdapterBackedSqlClient):
    """Extends the AdapterBackedSqlClient with the DDL methods necessary for test configuration and execution."""

    def create_table_from_dataframe(
        self,
        sql_table: SqlTable,
        df: pd.DataFrame,
        chunk_size: Optional[int] = None,
    ) -> None:
        """Create a table in the data warehouse containing the contents of the dataframe.

        Only used in tutorials and tests.

        Args:
            sql_table: The SqlTable object representing the table location to use
            df: The Pandas DataFrame object containing the column schema and data to load
            chunk_size: The number of rows to insert per transaction
        """
        logger.info(f"Creating table '{sql_table.sql}' from a DataFrame with {df.shape[0]} row(s)")
        start_time = time.time()
        with self._adapter.connection_named("MetricFlow_create_from_dataframe"):
            # Create table
            # update dtypes to convert None to NA in boolean columns.
            # This mirrors the SQLAlchemy schema detection logic in pandas.io.sql
            df = df.convert_dtypes()
            columns = df.columns
            columns_to_insert = []
            for i in range(len(df.columns)):
                # Format as "column_name column_type"
                columns_to_insert.append(
                    f"{columns[i]} {self._get_type_from_pandas_dtype(str(df[columns[i]].dtype).lower())}"
                )
            self._adapter.execute(
                f"CREATE TABLE IF NOT EXISTS {sql_table.sql} ({', '.join(columns_to_insert)})",
                auto_begin=True,
                fetch=False,
            )
            self._adapter.commit_if_has_connection()

            # Insert rows
            values = []
            for row in df.itertuples(index=False, name=None):
                cells = []
                for cell in row:
                    if pd.isnull(cell):
                        # use null keyword instead of isNA/None/etc.
                        cells.append("null")
                    elif type(cell) in [str, pd.Timestamp]:
                        # Wrap cell in quotes & escape existing single quotes
                        escaped_cell = self._quote_escape_value(str(cell))
                        cells.append(f"'{escaped_cell}'")
                    else:
                        cells.append(str(cell))

                values.append(f"({', '.join(cells)})")
                if chunk_size and len(values) == chunk_size:
                    value_string = ",\n".join(values)
                    self._adapter.execute(
                        f"INSERT INTO {sql_table.sql} VALUES {value_string}", auto_begin=True, fetch=False
                    )
                    values = []
            if values:
                value_string = ",\n".join(values)
                self._adapter.execute(
                    f"INSERT INTO {sql_table.sql} VALUES {value_string}", auto_begin=True, fetch=False
                )
            # Commit all insert transaction at once
            self._adapter.commit_if_has_connection()

        logger.info(f"Created table '{sql_table.sql}' from a DataFrame in {time.time() - start_time:.2f}s")

    def _get_type_from_pandas_dtype(self, dtype: str) -> str:
        """Helper method to get the engine-specific type value.

        The dtype dict here is non-exhaustive but should be adequate for our needs.
        """
        # TODO: add type handling for string/bool/bigint types for all engines
        if dtype == "string" or dtype == "object":
            if self.sql_engine_type is SqlEngine.DATABRICKS or self.sql_engine_type is SqlEngine.BIGQUERY:
                return "string"
            return "text"
        elif dtype == "boolean" or dtype == "bool":
            return "boolean"
        elif dtype == "int64":
            return "bigint"
        elif dtype == "float64":
            return self._sql_query_plan_renderer.expr_renderer.double_data_type
        elif dtype == "datetime64[ns]":
            return self._sql_query_plan_renderer.expr_renderer.timestamp_data_type
        else:
            raise ValueError(f"Encountered unexpected Pandas dtype ({dtype})!")

    def _quote_escape_value(self, value: str) -> str:
        """Escape single quotes in string-like values for create_table_from_dataframe.

        This is necessary because Databricks uses backslash as its escape character.
        We don't bother with the exhaustive switch here because we expect most engines
        to fit into the default single quote condition.
        """
        if self.sql_engine_type is SqlEngine.DATABRICKS or self.sql_engine_type is SqlEngine.BIGQUERY:
            return value.replace("'", "\\'")

        return value.replace("'", "''")

    def create_schema(self, schema_name: str) -> None:
        """Create the given schema in a data warehouse. Only used in tutorials and tests."""
        self.execute(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")

    def drop_schema(self, schema_name: str, cascade: bool = True) -> None:
        """Drop the given schema from the data warehouse. Only used in tests."""
        self.execute(f"DROP SCHEMA IF EXISTS {schema_name}{' CASCADE' if cascade else ''}")



from __future__ import annotations

import datetime
import logging
import os
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

import dateutil.parser
import pandas as pd
import yaml
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.implementations.base import FrozenBaseModel

from metricflow.dataflow.sql_table import SqlTable
from metricflow.specs.specs import hash_items
from metricflow.test.fixtures.sql_clients.ddl_sql_client import SqlClientWithDDLMethods

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlTableSnapshotHash:
    """A hash that can be used to compare whether a SQL table snapshot has changed."""

    str_value: str

    @staticmethod
    def create_from_hashes(hashes: Sequence[SqlTableSnapshotHash]) -> SqlTableSnapshotHash:  # noqa: D
        return SqlTableSnapshotHash(hash_items(tuple(one_hash.str_value for one_hash in hashes)))


class SqlTableColumnType(Enum):  # noqa: D
    STRING = "STRING"
    TIME = "TIME"
    FLOAT = "FLOAT"
    INT = "INT"
    BOOLEAN = "BOOLEAN"


class SqlTableColumnDefinition(FrozenBaseModel):
    """Pydantic class to help parse column definitions in a table snapshots that are defined in YAML."""

    # Pydantic feature to throw errors on extra fields.
    class Config:  # noqa: D
        extra = "forbid"

    name: str
    type: SqlTableColumnType


class SqlTableSnapshotTypeException(Exception):  # noqa: D
    pass


class SqlTableSnapshot(FrozenBaseModel):
    """Pydantic class to help parse table snapshots that are defined in YAML."""

    # Pydantic feature to throw errors on extra fields.
    class Config:  # noqa: D
        extra = "forbid"

    table_name: str
    column_definitions: Tuple[SqlTableColumnDefinition, ...]
    rows: Tuple[Tuple[Optional[str], ...], ...]
    file_path: Path

    @property
    def snapshot_hash(self) -> SqlTableSnapshotHash:
        """Return a hash that can be used to summarize the schema and data of the snapshot."""
        return SqlTableSnapshotHash(
            hash_items(
                (self.table_name,)
                + tuple(column_definition.name for column_definition in self.column_definitions)
                + tuple(column_definition.type.name for column_definition in self.column_definitions)
                + tuple(cell or "" for row in self.rows for cell in row)
            )
        )

    @staticmethod
    def _parse_bool_str(bool_str: str) -> bool:  # noqa: D
        if bool_str.lower() == "false":
            return False
        elif bool_str.lower() == "true":
            return True
        else:
            raise RuntimeError(f"Invalid string representation of a boolean: {bool_str}")

    @property
    def as_df(self) -> pd.DataFrame:  # noqa: D
        """Return this snapshot as represented by an equivalent dataframe."""
        # In the YAML files, all values are strings, but they need to be converted to defined type so that it can be
        # properly represented in a dataframe

        type_converted_rows = []
        for row in self.rows:
            type_converted_row: List[Union[str, datetime.datetime, int, float, bool, None]] = []
            for column_num, column_value in enumerate(row):
                column_type = self.column_definitions[column_num].type
                if column_value is None:
                    type_converted_row.append(None)
                elif column_type is SqlTableColumnType.STRING:
                    type_converted_row.append(column_value)
                elif column_type is SqlTableColumnType.TIME:
                    type_converted_row.append(dateutil.parser.parse(column_value))
                elif column_type is SqlTableColumnType.INT:
                    type_converted_row.append(int(column_value))
                elif column_type is SqlTableColumnType.FLOAT:
                    type_converted_row.append(float(column_value))
                elif column_type is SqlTableColumnType.BOOLEAN:
                    type_converted_row.append(SqlTableSnapshot._parse_bool_str(column_value))
                else:
                    assert_values_exhausted(column_type)
            type_converted_rows.append(type_converted_row)

        return pd.DataFrame(
            columns=[column_definition.name for column_definition in self.column_definitions],
            data=type_converted_rows,
        )


class SqlTableSnapshotLoader:
    """Loads a snapshot of a table into the SQL engine."""

    def __init__(self, ddl_sql_client: SqlClientWithDDLMethods, schema_name: str) -> None:  # noqa: D
        self._ddl_sql_client = ddl_sql_client
        self._schema_name = schema_name

    def load(self, table_snapshot: SqlTableSnapshot) -> None:  # noqa: D
        sql_table = SqlTable(schema_name=self._schema_name, table_name=table_snapshot.table_name)

        self._ddl_sql_client.create_table_from_dataframe(
            sql_table=sql_table,
            df=table_snapshot.as_df,
            # Without this set, the insert queries may be too long.
            chunk_size=500,
        )


class TableSnapshotException(Exception):  # noqa: D
    pass


class TableSnapshotParseException(TableSnapshotException):  # noqa: D
    pass


class SqlTableSnapshotRepository:
    """Stores table snapshots generated by parsing YAML files."""

    # The top level key in the YAML file
    TABLE_SNAPSHOT_DOCUMENT_KEY = "table_snapshot"

    def __init__(self, config_directory: Path) -> None:
        """Initializer.

        Args:
            config_directory: directory that should be searched for YAML files containing test cases.
        """
        self._config_directory = config_directory
        self._yaml_paths = SqlTableSnapshotRepository._find_all_yaml_file_paths(self._config_directory)
        self._table_snapshots: OrderedDict[str, SqlTableSnapshot] = OrderedDict()

        for file_path in self._yaml_paths:
            table_snapshots = SqlTableSnapshotRepository._parse_config_yaml(file_path)
            for table_snapshot in table_snapshots:
                if table_snapshot.table_name in self._table_snapshots:
                    raise ValueError(
                        f"Table with duplicate name found: {table_snapshot.table_name}. "
                        f"Conflicting paths: {table_snapshot.file_path} "
                        f"{self._table_snapshots[table_snapshot.table_name].file_path}"
                    )
                self._table_snapshots[table_snapshot.table_name] = table_snapshot

    @staticmethod
    def _parse_config_yaml(file_path: Path) -> Sequence[SqlTableSnapshot]:
        """Parse the YAML file at the given path into table snapshots."""
        results = []
        with open(file_path) as f:
            file_contents = f.read()
            for config_document in yaml.load_all(stream=file_contents, Loader=yaml.SafeLoader):
                # The config document can be None if there is nothing but white space between two `---`
                # this isn't really an issue, so lets just swallow it
                if config_document is None:
                    continue
                if not isinstance(config_document, dict):
                    raise TableSnapshotParseException(
                        f"Table snapshot YAML must be a dict. Got `{type(config_document)}`: {config_document}"
                    )

                keys = tuple(x for x in config_document.keys())
                if len(keys) != 1:
                    raise TableSnapshotParseException(
                        f"Table snapshot document should have one type of key, but this has {len(keys)}. "
                        f"Found keys: {keys} in {file_path}",
                    )

                # retrieve last top-level key as type
                document_type = next(iter(config_document.keys()))
                object_cfg = config_document[document_type]
                if document_type == SqlTableSnapshotRepository.TABLE_SNAPSHOT_DOCUMENT_KEY:
                    try:
                        results.append(SqlTableSnapshot(**object_cfg, file_path=Path(file_path)))
                    except Exception as e:
                        logger.exception(f"Error while parsing: {file_path}")
                        raise TableSnapshotParseException(f"Error while parsing: {file_path}") from e
                else:
                    raise TableSnapshotParseException(
                        f"Expected {SqlTableSnapshotRepository.TABLE_SNAPSHOT_DOCUMENT_KEY}, but got {document_type}"
                    )
        return results

    @staticmethod
    def _find_all_yaml_file_paths(directory: Path) -> Sequence[Path]:  # noqa: D
        """Recursively search through the given directory for YAML files."""
        yaml_file_paths = []

        for root, dirs, files in os.walk(directory):
            for file in files:
                if file.endswith(".yaml"):
                    yaml_file_paths.append(Path(root, file))

        return sorted(yaml_file_paths)

    @property
    def table_snapshots(self) -> Sequence[SqlTableSnapshot]:  # noqa: D
        # tuple(self._table_snapshots.values()) shows a type warning
        return sorted(
            tuple(table_snapshot for table_snapshot in self._table_snapshots.values()),
            key=lambda table_snapshot: table_snapshot.table_name,
        )


TABLE_SNAPSHOT_REPOSITORY = SqlTableSnapshotRepository(
    Path(os.path.dirname(__file__)),
)



from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field
from typing import List, Optional

from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from pydantic import ValidationError

from metricflow.sql_request.sql_request_attributes import SqlJsonTag, SqlRequestTagSet

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class CombinedSqlTags:
    """Groups system and extra tags for simplicity."""

    system_tags: SqlRequestTagSet = field(default_factory=SqlRequestTagSet)
    extra_tag: SqlJsonTag = field(default_factory=SqlJsonTag)


class SqlStatementCommentMetadata:
    """Helps to add a comment to SQL statements to encode metadata (e.g. tags).

    Added at the end as some engines remove leading comments:
    https://docs.snowflake.com/en/release-notes/2017-04.html#queries-leading-comments-removed-during-execution

    Example:
        SELECT 1
    ->
        -- MF_REQUEST_METADATA: {"tag_dict": {"MF_REQUEST_ID": "mf_rid__tmhulwkt"}}
        SELECT 1
    """

    _TAG_PREFIX = "-- MF_REQUEST_METADATA: "
    _EXTRA_TAG_PREFIX = "-- MF_EXTRA_TAGS: "

    @staticmethod
    def add_tag_metadata_as_comment(sql_statement: str, combined_tags: CombinedSqlTags) -> str:  # noqa: D
        if combined_tags.system_tags.tags:
            sql_statement = (
                sql_statement + "\n" + SqlStatementCommentMetadata._TAG_PREFIX + combined_tags.system_tags.json()
            )

        if combined_tags.extra_tag.json_dict:
            serialized_json: Optional[str] = None
            try:
                serialized_json = json.dumps(combined_tags.extra_tag.json_dict)
            except Exception:
                logger.exception(f"Not including extra tag that couldn't be serialized: {combined_tags.extra_tag}\n")

            if serialized_json:
                sql_statement = sql_statement + "\n" + SqlStatementCommentMetadata._EXTRA_TAG_PREFIX + serialized_json

        return sql_statement

    @staticmethod
    def parse_tag_metadata_in_comments(sql_statement: str) -> CombinedSqlTags:  # noqa: D
        tag_sets: List[SqlRequestTagSet] = []
        extra_tags: List[SqlJsonTag] = []
        for line in sql_statement.split("\n"):
            if line.startswith(SqlStatementCommentMetadata._TAG_PREFIX):
                try:
                    json_str = line[len(SqlStatementCommentMetadata._TAG_PREFIX) :]
                    tag_sets.append(SqlRequestTagSet.parse_raw(json_str))
                except ValidationError:
                    logger.exception(f"Unable to parse tag metadata from line: {line}")

            if line.startswith(SqlStatementCommentMetadata._EXTRA_TAG_PREFIX):
                try:
                    json_str = line[len(SqlStatementCommentMetadata._EXTRA_TAG_PREFIX) :]
                    extra_tags.append(SqlJsonTag(json.loads(json_str)))
                except ValidationError:
                    logger.exception(f"Unable to parse extra tag metadata from line: {line}")

        if len(tag_sets) > 1:
            logger.error(
                f"Got multiple tag sets from parsing comments:\n"
                f"{pformat_big_objects(tag_sets)}\n"
                f"Using the first one."
            )

        if len(extra_tags) > 1:
            logger.error(
                f"Got multiple extra tags from parsing comments:\n"
                f"{pformat_big_objects(extra_tags)}\n"
                f"Using the first one."
            )

        return CombinedSqlTags(
            system_tags=tag_sets[0] if len(tag_sets) > 0 else SqlRequestTagSet(),
            extra_tag=extra_tags[0] if len(extra_tags) > 0 else SqlJsonTag(),
        )



from __future__ import annotations

import logging
import os
from pathlib import Path

import pytest

from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.sql_clients.ddl_sql_client import SqlClientWithDDLMethods
from metricflow.test.source_schema_tools import create_tables_listed_in_table_snapshot_repository
from metricflow.test.table_snapshot.table_snapshots import (
    SqlTableSnapshotRepository,
)

logger = logging.getLogger(__name__)


# Prefer to use the fixture, but there are some cases where fixtures aren't available. e.g. when defining a
# parameterized test.
CONFIGURED_SOURCE_TABLE_SNAPSHOT_REPOSITORY = SqlTableSnapshotRepository(
    Path(os.path.dirname(__file__)).joinpath("source_table_snapshots")
)


@pytest.fixture(scope="session")
def source_table_snapshot_repository() -> SqlTableSnapshotRepository:  # noqa: D
    return CONFIGURED_SOURCE_TABLE_SNAPSHOT_REPOSITORY


@pytest.fixture(scope="session", autouse=True)
def check_time_spine_source(
    mf_test_session_state: MetricFlowTestSessionState,
    source_table_snapshot_repository: SqlTableSnapshotRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Check that the time spine source follows the definition in the table snapshot.

    The time spine table is defined in a table snapshot YAML file and is restored to the source schema based on that
    definition. The definition in the YAML should align with the definition in the time_spine_source fixture.
    """
    time_spine_source = simple_semantic_manifest_lookup.time_spine_source
    assert (
        time_spine_source.schema_name == mf_test_session_state.mf_source_schema
    ), "The time spine source table should be in the source schema"

    time_spine_snapshot_candidates = tuple(
        snapshot
        for snapshot in source_table_snapshot_repository.table_snapshots
        if snapshot.table_name == time_spine_source.table_name
    )

    assert len(time_spine_snapshot_candidates) == 1, (
        f"Did not get exactly one table snapshot matching the time_spine_source table name. "
        f"Got: {time_spine_snapshot_candidates}"
    )

    time_spine_snapshot = time_spine_snapshot_candidates[0]

    assert len(time_spine_snapshot.column_definitions) == 1
    time_column = time_spine_snapshot.column_definitions[0]
    assert time_column.name == time_spine_source.time_column_name


@pytest.fixture(scope="session")
def create_source_tables(
    mf_test_session_state: MetricFlowTestSessionState,
    ddl_sql_client: SqlClientWithDDLMethods,
    source_table_snapshot_repository: SqlTableSnapshotRepository,
) -> None:
    """Creates all tables that should be in the source schema.

    If a table with a given name already exists in the source schema, it's assumed to have the expected schema / data.
    """
    if mf_test_session_state.use_persistent_source_schema:
        logger.info(
            "This session was configured to use a persistent source schema, so this fixture won't create new tables. "
            "See populate_source_schema() for more details."
        )
        return

    create_tables_listed_in_table_snapshot_repository(
        ddl_sql_client=ddl_sql_client,
        schema_name=mf_test_session_state.mf_source_schema,
        table_snapshot_repository=source_table_snapshot_repository,
    )



from __future__ import annotations

from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_execution import DataflowToExecutionPlanConverter
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs.specs import (
    DimensionSpec,
    EntityReference,
    MetricFlowQuerySpec,
    MetricSpec,
    TimeDimensionSpec,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.snapshot_utils import assert_execution_plan_text_equal


def make_execution_plan_converter(  # noqa: D
    semantic_manifest_lookup: SemanticManifestLookup,
    sql_client: SqlClient,
) -> DataflowToExecutionPlanConverter:
    return DataflowToExecutionPlanConverter(
        sql_plan_converter=DataflowToSqlQueryPlanConverter(
            column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup),
            semantic_manifest_lookup=semantic_manifest_lookup,
        ),
        sql_plan_renderer=DefaultSqlQueryPlanRenderer(),
        sql_client=sql_client,
    )


def test_joined_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_plan_builder: DataflowPlanBuilder,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    sql_client: SqlClient,
) -> None:
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(MetricSpec(element_name="bookings"),),
            dimension_specs=(
                DimensionSpec(
                    element_name="is_instant",
                    entity_links=(),
                ),
                DimensionSpec(
                    element_name="country_latest",
                    entity_links=(EntityReference("listing"),),
                ),
            ),
        )
    )

    to_execution_plan_converter = make_execution_plan_converter(
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
        sql_client=sql_client,
    )

    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )


def test_small_combined_metrics_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    dataflow_plan_builder: DataflowPlanBuilder,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(
                MetricSpec(element_name="bookings"),
                MetricSpec(element_name="booking_value"),
            ),
            dimension_specs=(
                DimensionSpec(
                    element_name="is_instant",
                    entity_links=(),
                ),
            ),
        )
    )

    to_execution_plan_converter = make_execution_plan_converter(
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
        sql_client=sql_client,
    )
    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )


def test_combined_metrics_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    dataflow_plan_builder: DataflowPlanBuilder,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(
                MetricSpec(element_name="bookings"),
                MetricSpec(element_name="instant_bookings"),
                MetricSpec(element_name="booking_value"),
            ),
            dimension_specs=(
                DimensionSpec(
                    element_name="is_instant",
                    entity_links=(),
                ),
            ),
            time_dimension_specs=(TimeDimensionSpec(element_name="ds", entity_links=()),),
        )
    )

    to_execution_plan_converter = make_execution_plan_converter(
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
        sql_client=sql_client,
    )
    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )


def test_multihop_joined_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    multihop_dataflow_plan_builder: DataflowPlanBuilder,
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
    sql_client: SqlClient,
) -> None:
    """Tests a plan getting a measure and a joined dimension."""
    dataflow_plan = multihop_dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(MetricSpec(element_name="txn_count"),),
            dimension_specs=(
                DimensionSpec(
                    element_name="customer_name",
                    entity_links=(
                        EntityReference(element_name="account_id"),
                        EntityReference(element_name="customer_id"),
                    ),
                ),
            ),
        )
    )

    to_execution_plan_converter = DataflowToExecutionPlanConverter(
        sql_plan_converter=DataflowToSqlQueryPlanConverter(
            column_association_resolver=DunderColumnAssociationResolver(multi_hop_join_semantic_manifest_lookup),
            semantic_manifest_lookup=multi_hop_join_semantic_manifest_lookup,
        ),
        sql_plan_renderer=DefaultSqlQueryPlanRenderer(),
        sql_client=sql_client,
    )

    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )



from __future__ import annotations

import difflib
import logging
import os
import re
import webbrowser
from typing import Any, Callable, List, Optional, Tuple, TypeVar

import tabulate
from _pytest.fixtures import FixtureRequest
from dbt_semantic_interfaces.pretty_print import pformat_big_objects

from metricflow.dag.mf_dag import MetricFlowDag
from metricflow.dataflow.dataflow_plan import DataflowPlan
from metricflow.dataflow.dataflow_plan_to_text import dataflow_plan_as_text
from metricflow.execution.execution_plan import ExecutionPlan
from metricflow.execution.execution_plan_to_text import execution_plan_to_text
from metricflow.model.semantics.linkable_spec_resolver import LinkableElementSet
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs.specs import InstanceSpecSet
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState

logger = logging.getLogger(__name__)


# In plan outputs, replace strings that vary from run to run with this so that comparisons can be made
# consistently.
PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS = "*"


def make_schema_replacement_function(system_schema: str, source_schema: str) -> Callable[[str], str]:
    """Generates a function to replace schema names in test outputs."""

    # The schema of the warehouse used in tests changes from run to run, so don't compare those.
    def replacement_function(text: str) -> str:  # noqa: D
        # Replace with a string of the same length so that indents are preserved.
        text = text.replace(source_schema, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(source_schema))
        # Same with the MetricFlow system schema.
        return text.replace(system_schema, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(system_schema))

    return replacement_function


def replace_dataset_id_hash(text: str) -> str:
    """Replaces data set ID hashes for primed semantic models.

    The data set ID hash changes from run to run because it's based on the DW schema in the semantic model, which changes
    run to run.
    """
    pattern = re.compile(r"'[a-zA-Z0-9_]+__[a-zA-Z0-9_]+__(?P<hash>[a-zA-Z0-9_]+)'")
    while True:
        match = pattern.search(text)
        if match:
            data_set_id_hash = match.group("hash")
            # Replace with the same length to preserve indents
            text = text.replace(data_set_id_hash, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(data_set_id_hash))
        else:
            break
    return text


def snapshot_path_prefix(
    request: FixtureRequest,
    snapshot_group: str,
    snapshot_id: str,
    additional_sub_directories: Tuple[str, ...] = (),
) -> str:
    """Returns a path prefix that can be used to build filenames for files associated with the snapshot.

    The snapshot prefix is generated from the name of the test file, the name of the test, name of the snapshot class,
    and the name of the snapshot.

    e.g.
    .../snapshots/test_file.py/DataflowPlan/test_name__plan1

    which can be used to construct paths like

    .../snapshots/test_file.py/DataflowPlan/test_name__plan1.xml
    .../snapshots/test_file.py/DataflowPlan/test_name__plan1.svg
    """
    test_name = request.node.name
    snapshot_file_name = test_name + "__" + snapshot_id
    path_items: List[str] = []

    test_file_path_items = os.path.normpath(request.node.fspath).split(os.sep)
    test_file_name = test_file_path_items[-1]
    # Default to where this is defined, but use more appropriate directories if found.
    test_directory_root_index = -1
    for i, path_item in enumerate(test_file_path_items):
        if path_item in ("test", "metricflow", "metricflow_extensions"):
            test_directory_root_index = i + 1

    path_to_store_snapshots = os.sep.join(test_file_path_items[:test_directory_root_index])
    path_items.extend([path_to_store_snapshots, "snapshots", test_file_name, snapshot_group])

    if additional_sub_directories:
        path_items.extend(additional_sub_directories)
    path_items.append(snapshot_file_name)

    return os.path.abspath(os.path.join(*path_items))


PlanT = TypeVar("PlanT", bound=MetricFlowDag)


def _exclude_lines_matching_regex(file_contents: str, exclude_line_regex: str) -> str:
    """Removes lines from file contents if the line matches exclude_regex."""
    compiled_regex = re.compile(exclude_line_regex)
    return "\n".join([line for line in file_contents.split("\n") if not compiled_regex.match(line)])


def assert_plan_snapshot_text_equal(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    plan: PlanT,
    plan_snapshot_text: str,
    plan_snapshot_file_extension: str = ".xml",
    exclude_line_regex: Optional[str] = None,
    incomparable_strings_replacement_function: Optional[Callable[[str], str]] = None,
    additional_sub_directories_for_snapshots: Tuple[str, ...] = (),
) -> None:
    """Checks if the given plan text is equal to the one that's saved for comparison.

    * The location of the file is automatically generated based on the test and the plan's ID.
    * This may create a new saved plan file or overwrite the existing one, depending on the configuration.
    * replace_incomparable_strings is used to replace strings in the plan text before comparison. Useful for making
      plans consistent when there are strings that vary between runs and shouldn't be compared.
    * additional_sub_directories_for_snapshots is used to specify additional sub-directories (in the automatically
      generated directory) where plan outputs should reside.
    """
    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id=plan.__class__.__name__,
        snapshot_id=plan.dag_id,
        snapshot_text=plan_snapshot_text,
        snapshot_file_extension=plan_snapshot_file_extension,
        exclude_line_regex=exclude_line_regex,
        incomparable_strings_replacement_function=incomparable_strings_replacement_function,
        additional_sub_directories_for_snapshots=additional_sub_directories_for_snapshots,
    )


def assert_snapshot_text_equal(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    group_id: str,
    snapshot_id: str,
    snapshot_text: str,
    snapshot_file_extension: str,
    exclude_line_regex: Optional[str] = None,
    incomparable_strings_replacement_function: Optional[Callable[[str], str]] = None,
    additional_sub_directories_for_snapshots: Tuple[str, ...] = (),
) -> None:
    """Similar to assert_plan_snapshot_text_equal(), but with more controls on how the snapshot paths are generated."""
    file_path = (
        snapshot_path_prefix(
            request=request,
            snapshot_group=group_id,
            snapshot_id=snapshot_id,
            additional_sub_directories=additional_sub_directories_for_snapshots,
        )
        + snapshot_file_extension
    )

    if incomparable_strings_replacement_function is not None:
        snapshot_text = incomparable_strings_replacement_function(snapshot_text)

    # If we are in overwrite mode, make a new plan:
    if mf_test_session_state.overwrite_snapshots:
        # Create parent directory for the plan text files.
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as snapshot_text_file:
            snapshot_text_file.write(snapshot_text)
            # Add a new line at the end of the file so that PRSs don't show the "no newline" symbol on Github.
            snapshot_text_file.write("\n")

    # Throw an exception if the plan is not there.
    if not os.path.exists(file_path):
        raise FileNotFoundError(
            f"Could not find snapshot file at path {file_path}. Re-run with --overwrite-snapshots and check git status "
            "to see what's new."
        )

    if mf_test_session_state.display_plans:
        if not mf_test_session_state.overwrite_snapshots:
            logger.warning(f"Not overwriting snapshots, so displaying existing snapshot at {file_path}")

        if mf_test_session_state.plans_displayed >= mf_test_session_state.max_plans_displayed:
            raise RuntimeError(
                f"Can't display snapshot - hit limit of "
                f"{mf_test_session_state.max_plans_displayed} "
                f"plans displayed."
            )
        webbrowser.open("file://" + file_path)
        mf_test_session_state.plans_displayed += 1

    # Read the existing plan from the file and compare with the actual plan
    with open(file_path, "r") as snapshot_text_file:
        # Remove the newline that was added from above.
        expected_snapshot_text = snapshot_text_file.read().rstrip()

        if exclude_line_regex:
            # Filter out lines that should be ignored.
            expected_snapshot_text = _exclude_lines_matching_regex(
                file_contents=expected_snapshot_text, exclude_line_regex=exclude_line_regex
            )
            snapshot_text = _exclude_lines_matching_regex(
                file_contents=snapshot_text, exclude_line_regex=exclude_line_regex
            )
        # pytest should show a detailed diff with "assert actual_modified == expected_modified", but it's not, so doing
        # this instead.
        if snapshot_text != expected_snapshot_text:
            differ = difflib.Differ()
            diff = differ.compare(expected_snapshot_text.splitlines(), snapshot_text.splitlines())
            assert False, f"Snapshot from {file_path} does not match. Diff from expected to actual:\n" + "\n".join(diff)


def assert_execution_plan_text_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    execution_plan: ExecutionPlan,
) -> None:
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=execution_plan,
        plan_snapshot_text=execution_plan_to_text(execution_plan),
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=mf_test_session_state.mf_system_schema,
            source_schema=mf_test_session_state.mf_source_schema,
        ),
        additional_sub_directories_for_snapshots=(sql_client.sql_engine_type.value,),
    )


def assert_dataflow_plan_text_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_plan: DataflowPlan,
    sql_client: SqlClient,
) -> None:
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=dataflow_plan,
        plan_snapshot_text=dataflow_plan_as_text(dataflow_plan),
        incomparable_strings_replacement_function=replace_dataset_id_hash,
        additional_sub_directories_for_snapshots=(sql_client.sql_engine_type.value,),
    )


def assert_object_snapshot_equal(  # type: ignore[misc]
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    obj_id: str,
    obj: Any,
) -> None:
    """For tests to compare large objects, this can be used to snapshot a text representation of the object."""
    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id=obj.__class__.__name__,
        snapshot_id=obj_id,
        snapshot_text=pformat_big_objects(obj),
        snapshot_file_extension=".txt",
    )


def assert_linkable_element_set_snapshot_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    set_id: str,
    linkable_element_set: LinkableElementSet,
) -> None:
    headers = ("Semantic Model", "Entity Links", "Name", "Time Granularity", "Properties")
    rows = []
    for linkable_dimension_iterable in linkable_element_set.path_key_to_linkable_dimensions.values():
        for linkable_dimension in linkable_dimension_iterable:
            rows.append(
                (
                    # Checking a limited set of fields as the result is large due to the paths in the object.
                    linkable_dimension.semantic_model_origin.semantic_model_name,
                    tuple(entity_link.element_name for entity_link in linkable_dimension.entity_links),
                    linkable_dimension.element_name,
                    linkable_dimension.time_granularity.name if linkable_dimension.time_granularity is not None else "",
                    sorted(
                        linkable_element_property.name for linkable_element_property in linkable_dimension.properties
                    ),
                )
            )

    for linkable_entity_iterable in linkable_element_set.path_key_to_linkable_entities.values():
        for linkable_entity in linkable_entity_iterable:
            rows.append(
                (
                    # Checking a limited set of fields as the result is large due to the paths in the object.
                    linkable_entity.semantic_model_origin.semantic_model_name,
                    tuple(entity_link.element_name for entity_link in linkable_entity.entity_links),
                    linkable_entity.element_name,
                    "",
                    sorted(linkable_element_property.name for linkable_element_property in linkable_entity.properties),
                )
            )
    assert_object_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        obj_id=set_id,
        obj=tabulate.tabulate(headers=headers, tabular_data=sorted(rows)),
    )


def assert_spec_set_snapshot_equal(  # noqa: D
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, set_id: str, spec_set: InstanceSpecSet
) -> None:
    assert_object_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        obj_id=set_id,
        obj=sorted(spec.qualified_name for spec in spec_set.all_specs),
    )



from __future__ import annotations

from abc import abstractmethod
from typing import Optional, Protocol

from pandas import DataFrame

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient


class SqlClientWithDDLMethods(SqlClient, Protocol):
    """SqlClient protocol with DDL methods enabled.

    The core MetricFlow SqlClient is effectively a read-only construct, as MetricFlow is meant to compile and run SQL
    for metric queries, not manage downstream database state.

    However, it can be useful to have a SqlClient that can do DDL. In this case, we use these methods in our warehouse
    engine integration tests for setup and teardown. All of the SqlClients we use in our tests must implement this
    protocol, or else the integration tests cannot be run.
    """

    @abstractmethod
    def create_table_from_dataframe(
        self,
        sql_table: SqlTable,
        df: DataFrame,
        chunk_size: Optional[int] = None,
    ) -> None:
        """Creates a table and populates it with the contents of the dataframe.

        Args:
            sql_table: The SqlTable metadata of the table to create
            df: The Pandas DataFrame with the contents of the target table
            chunk_size: The number of rows to write per query
        """
        raise NotImplementedError

    @abstractmethod
    def create_schema(self, schema_name: str) -> None:
        """Create the given schema if it doesn't already exist."""
        raise NotImplementedError

    @abstractmethod
    def drop_schema(self, schema_name: str, cascade: bool) -> None:  # noqa: D
        """Drop the given schema if it exists. If cascade is set, drop the tables in the schema first."""
        raise NotImplementedError



from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from typing import Optional

import pandas as pd
from dbt_semantic_interfaces.pretty_print import pformat_big_objects

from metricflow.dataflow.sql_table import SqlTable
from metricflow.formatting import indent_log_line
from metricflow.protocols.sql_client import (
    SqlClient,
)
from metricflow.random_id import random_id
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_request.sql_request_attributes import SqlJsonTag, SqlRequestId, SqlRequestTagSet
from metricflow.sql_request.sql_statement_metadata import CombinedSqlTags, SqlStatementCommentMetadata

logger = logging.getLogger(__name__)


class SqlClientException(Exception):
    """Raised when an interaction with the SQL engine has an error."""

    pass


class BaseSqlClientImplementation(ABC, SqlClient):
    """Abstract implementation that other SQL clients are based on."""

    @staticmethod
    def _format_run_query_log_message(statement: str, sql_bind_parameters: SqlBindParameters) -> str:
        message = f"Running query:\n\n{indent_log_line(statement)}"
        if len(sql_bind_parameters.param_dict) > 0:
            message += (
                f"\n"
                f"\n"
                f"with parameters:\n"
                f"\n"
                f"{indent_log_line(pformat_big_objects(sql_bind_parameters.param_dict))}"
            )
        return message

    def query(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> pd.DataFrame:
        """Query statement; result expected to be data which will be returned as a DataFrame.

        Args:
            stmt: The SQL query statement to run. This should produce output via a SELECT
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
        """
        start = time.time()
        request_id = SqlRequestId(f"mf_rid__{random_id()}")
        combined_tags = BaseSqlClientImplementation._consolidate_tags(json_tags=extra_tags, request_id=request_id)
        statement = SqlStatementCommentMetadata.add_tag_metadata_as_comment(
            sql_statement=stmt, combined_tags=combined_tags
        )
        logger.info(BaseSqlClientImplementation._format_run_query_log_message(statement, sql_bind_parameters))
        df = self._engine_specific_query_implementation(
            stmt=statement,
            bind_params=sql_bind_parameters,
            system_tags=combined_tags.system_tags,
            extra_tags=combined_tags.extra_tag,
        )
        if not isinstance(df, pd.DataFrame):
            raise RuntimeError(f"Expected query to return a DataFrame, got {type(df)}")
        stop = time.time()
        logger.info(f"Finished running the query in {stop - start:.2f}s with {df.shape[0]} row(s) returned")
        return df

    def execute(  # noqa: D
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> None:
        start = time.time()
        request_id = SqlRequestId(f"mf_rid__{random_id()}")
        combined_tags = BaseSqlClientImplementation._consolidate_tags(json_tags=extra_tags, request_id=request_id)
        statement = SqlStatementCommentMetadata.add_tag_metadata_as_comment(
            sql_statement=stmt, combined_tags=combined_tags
        )
        logger.info(BaseSqlClientImplementation._format_run_query_log_message(statement, sql_bind_parameters))
        self._engine_specific_execute_implementation(
            stmt=statement,
            bind_params=sql_bind_parameters,
            system_tags=combined_tags.system_tags,
            extra_tags=combined_tags.extra_tag,
        )
        stop = time.time()
        logger.info(f"Finished running the query in {stop - start:.2f}s")
        return None

    def dry_run(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Dry run statement; checks that the 'stmt' is queryable. Returns None. Raises an exception if the 'stmt' isn't queryable.

        Args:
            stmt: The SQL query statement to dry run.
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
        """
        start = time.time()
        logger.info(
            f"Running dry_run of:"
            f"\n\n{indent_log_line(stmt)}\n"
            + (f"\nwith parameters: {dict(sql_bind_parameters.param_dict)}" if sql_bind_parameters.param_dict else "")
        )
        results = self._engine_specific_dry_run_implementation(stmt, sql_bind_parameters)
        stop = time.time()
        logger.info(f"Finished running the dry_run in {stop - start:.2f}s")
        return results

    @abstractmethod
    def _engine_specific_query_implementation(
        self,
        stmt: str,
        bind_params: SqlBindParameters,
        system_tags: SqlRequestTagSet = SqlRequestTagSet(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> pd.DataFrame:
        """Sub-classes should implement this to query the engine."""
        pass

    @abstractmethod
    def _engine_specific_execute_implementation(
        self,
        stmt: str,
        bind_params: SqlBindParameters,
        system_tags: SqlRequestTagSet = SqlRequestTagSet(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> None:
        """Sub-classes should implement this to execute a statement that doesn't return results."""
        pass

    @abstractmethod
    def _engine_specific_dry_run_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:
        """Sub-classes should implement this to check a query will run successfully without actually running the query."""
        pass

    @abstractmethod
    def create_table_from_dataframe(  # noqa: D
        self,
        sql_table: SqlTable,
        df: pd.DataFrame,
        chunk_size: Optional[int] = None,
    ) -> None:
        pass

    def create_schema(self, schema_name: str) -> None:  # noqa: D
        self.execute(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")

    def drop_schema(self, schema_name: str, cascade: bool = True) -> None:  # noqa: D
        self.execute(f"DROP SCHEMA IF EXISTS {schema_name}{' CASCADE' if cascade else ''}")

    def close(self) -> None:  # noqa: D
        pass

    def render_bind_parameter_key(self, bind_parameter_key: str) -> str:
        """Wrap execution parameter key with syntax accepted by engine."""
        return f":{bind_parameter_key}"

    @staticmethod
    def _consolidate_tags(json_tags: SqlJsonTag, request_id: SqlRequestId) -> CombinedSqlTags:
        """Consolidates json tags and request ID into a single set of tags."""
        return CombinedSqlTags(
            system_tags=SqlRequestTagSet().add_request_id(request_id=request_id),
            extra_tag=json_tags,
        )



from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Set, Tuple

import pandas as pd
from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from dbt_semantic_interfaces.references import (
    EntityReference,
    MeasureReference,
    MetricReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import ReadSqlSourceNode
from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.specs.specs import (
    TimeDimensionSpec,
)
from metricflow.time.date_part import DatePart
from metricflow.time.time_granularity import (
    adjust_to_end_of_period,
    adjust_to_start_of_period,
    is_period_end,
    is_period_start,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PartialTimeDimensionSpec:
    """Similar to TimeDimensionSpec, but with an unknown time granularity.

    This is used to represent a time dimension spec from the user before the granularity is figured out.
    """

    element_name: str
    entity_links: Tuple[EntityReference, ...]
    date_part: Optional[DatePart] = None


@dataclass(frozen=True)
class LocalTimeDimensionGranularityKey:
    """Used to associate a measure and a local time dimension without being specific about a granularity."""

    measure_reference: MeasureReference
    local_time_dimension_reference: TimeDimensionReference


class TimeGranularitySolver:
    """Figures out time granularities for metrics.

    1. Given the list of metrics, what's the smallest time granularity that's common across those metrics?
    2. Given the list of metrics and the dimensions that the user specified, does the primary time dimension need to be
    converted into another granularity? e.g. if the metrics have different time granularities and smallest time
    granularity for the given metrics is MONTH and the user specified ds, then we need to convert ds to ds__month.
    3. Also checks to see if the time granularity expressed in dimensions is valid. e.g. if a metric has a time
    granularity of MONTH, throw an exception if ds__day is in the requested dimensions.
    """

    def __init__(  # noqa: D
        self,
        semantic_manifest_lookup: SemanticManifestLookup,
        node_output_resolver: DataflowPlanNodeOutputDataSetResolver,
        read_nodes: Sequence[ReadSqlSourceNode],
    ) -> None:
        self._semantic_manifest_lookup = semantic_manifest_lookup
        self._time_dimension_names_to_supported_granularities: Dict[str, Set[TimeGranularity]] = defaultdict(set)
        for read_node in read_nodes:
            output_data_set = node_output_resolver.get_output_data_set(read_node)
            for time_dimension_instance in output_data_set.instance_set.time_dimension_instances:
                if time_dimension_instance.spec.date_part:
                    continue
                granularity_free_qualified_name = StructuredLinkableSpecName.from_name(
                    time_dimension_instance.spec.qualified_name
                ).granularity_free_qualified_name
                self._time_dimension_names_to_supported_granularities[granularity_free_qualified_name].add(
                    time_dimension_instance.spec.time_granularity
                )

    def validate_time_granularity(
        self, metric_references: Sequence[MetricReference], time_dimension_specs: Sequence[TimeDimensionSpec]
    ) -> None:
        """Check that the granularity specified for time dimensions is valid with respect to the metrics.

        e.g. throw an error if "ds__week" is specified for a metric with a time granularity of MONTH.
        """
        if not metric_references:
            return None

        valid_group_by_elements = self._semantic_manifest_lookup.metric_lookup.linkable_set_for_metrics(
            metric_references=metric_references,
        )

        for time_dimension_spec in time_dimension_specs:
            match_found = False
            for path_key in valid_group_by_elements.path_key_to_linkable_dimensions:
                if (
                    path_key.element_name == time_dimension_spec.element_name
                    and (path_key.entity_links == time_dimension_spec.entity_links)
                    and path_key.time_granularity == time_dimension_spec.time_granularity
                ):
                    match_found = True
                    break
            if not match_found:
                raise RequestTimeGranularityException(
                    f"{time_dimension_spec} is not valid for querying {metric_references}"
                )

    def resolve_granularity_for_partial_time_dimension_specs(
        self,
        metric_references: Sequence[MetricReference],
        partial_time_dimension_specs: Sequence[PartialTimeDimensionSpec],
    ) -> Dict[PartialTimeDimensionSpec, TimeDimensionSpec]:
        """Figure out the lowest granularity possible for the partially specified time dimension specs.

        Returns a dictionary that maps how the partial time dimension spec should be turned into a time dimension spec.
        """
        result: Dict[PartialTimeDimensionSpec, TimeDimensionSpec] = {}

        for partial_time_dimension_spec in partial_time_dimension_specs:
            minimum_time_granularity = self.find_minimum_granularity_for_partial_time_dimension_spec(
                partial_time_dimension_spec=partial_time_dimension_spec, metric_references=metric_references
            )
            result[partial_time_dimension_spec] = TimeDimensionSpec(
                element_name=partial_time_dimension_spec.element_name,
                entity_links=partial_time_dimension_spec.entity_links,
                time_granularity=minimum_time_granularity,
                date_part=partial_time_dimension_spec.date_part,
            )
        return result

    def find_minimum_granularity_for_partial_time_dimension_spec(
        self, partial_time_dimension_spec: PartialTimeDimensionSpec, metric_references: Sequence[MetricReference]
    ) -> TimeGranularity:
        """Find minimum granularity allowed for time dimension when queried with given metrics."""
        minimum_time_granularity: Optional[TimeGranularity] = None

        if metric_references:
            valid_group_by_elements = self._semantic_manifest_lookup.metric_lookup.linkable_set_for_metrics(
                metric_references=metric_references,
            )
            for path_key in valid_group_by_elements.path_key_to_linkable_dimensions:
                if (
                    path_key.element_name == partial_time_dimension_spec.element_name
                    and path_key.entity_links == partial_time_dimension_spec.entity_links
                    and path_key.time_granularity is not None
                ):
                    minimum_time_granularity = (
                        path_key.time_granularity
                        if minimum_time_granularity is None
                        else min(minimum_time_granularity, path_key.time_granularity)
                    )
            if not minimum_time_granularity:
                raise RequestTimeGranularityException(
                    f"Unable to resolve the time dimension spec for {partial_time_dimension_spec}. "
                    f"Valid group by elements are:\n"
                    f"{pformat_big_objects([spec.qualified_name for spec in valid_group_by_elements.as_spec_set.as_tuple])}"
                )
        else:
            granularity_free_qualified_name = StructuredLinkableSpecName(
                entity_link_names=tuple(
                    [entity_link.element_name for entity_link in partial_time_dimension_spec.entity_links]
                ),
                element_name=partial_time_dimension_spec.element_name,
            ).granularity_free_qualified_name

            supported_granularities = self._time_dimension_names_to_supported_granularities.get(
                granularity_free_qualified_name
            )
            if not supported_granularities:
                raise RequestTimeGranularityException(
                    f"Unable to resolve the time dimension spec for {partial_time_dimension_spec}. "
                )
            minimum_time_granularity = min(
                self._time_dimension_names_to_supported_granularities[granularity_free_qualified_name]
            )

        return minimum_time_granularity

    def adjust_time_range_to_granularity(
        self, time_range_constraint: TimeRangeConstraint, time_granularity: TimeGranularity
    ) -> TimeRangeConstraint:
        """Change the time range so that the ends are at the ends of the appropriate time granularity windows.

        e.g. [2020-01-15, 2020-2-15] with MONTH granularity -> [2020-01-01, 2020-02-29]
        """
        constraint_start = time_range_constraint.start_time
        constraint_end = time_range_constraint.end_time

        start_ts = pd.Timestamp(time_range_constraint.start_time)
        if not is_period_start(time_granularity, start_ts):
            constraint_start = adjust_to_start_of_period(time_granularity, start_ts).to_pydatetime()

        end_ts = pd.Timestamp(time_range_constraint.end_time)
        if not is_period_end(time_granularity, end_ts):
            constraint_end = adjust_to_end_of_period(time_granularity, end_ts).to_pydatetime()

        if constraint_start < TimeRangeConstraint.ALL_TIME_BEGIN():
            constraint_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        if constraint_end > TimeRangeConstraint.ALL_TIME_END():
            constraint_end = TimeRangeConstraint.ALL_TIME_END()

        return TimeRangeConstraint(start_time=constraint_start, end_time=constraint_end)


class RequestTimeGranularityException(Exception):
    """Raised when a query is requesting a time granularity that's not possible for the given metrics."""

    pass



from __future__ import annotations

from typing import Sequence

import pytest
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.specs.specs import (
    DimensionSpec,
    EntityReference,
    EntitySpec,
    InstanceSpec,
    InstanceSpecSet,
    LinkableInstanceSpec,
    LinklessEntitySpec,
    MeasureSpec,
    MetricSpec,
    TimeDimensionSpec,
)


@pytest.fixture
def dimension_spec() -> DimensionSpec:  # noqa: D
    return DimensionSpec(
        element_name="platform",
        entity_links=(
            EntityReference(element_name="user_id"),
            EntityReference(element_name="device_id"),
        ),
    )


@pytest.fixture
def time_dimension_spec() -> TimeDimensionSpec:  # noqa: D
    return TimeDimensionSpec(
        element_name="signup_ts",
        entity_links=(EntityReference(element_name="user_id"),),
        time_granularity=TimeGranularity.DAY,
    )


@pytest.fixture
def entity_spec() -> EntitySpec:  # noqa: D
    return EntitySpec(
        element_name="user_id",
        entity_links=(EntityReference(element_name="listing_id"),),
    )


def test_merge_specs(dimension_spec: DimensionSpec, entity_spec: EntitySpec) -> None:
    """Tests InstanceSpec.merge()."""
    assert InstanceSpec.merge([dimension_spec], [entity_spec]) == [dimension_spec, entity_spec]


def test_dimension_without_first_entity_link(dimension_spec: DimensionSpec) -> None:  # noqa: D
    assert dimension_spec.without_first_entity_link == DimensionSpec(
        element_name="platform", entity_links=(EntityReference(element_name="device_id"),)
    )


def test_dimension_without_entity_links(dimension_spec: DimensionSpec) -> None:  # noqa: D
    assert dimension_spec.without_entity_links == DimensionSpec(element_name="platform", entity_links=())


def test_time_dimension_without_first_entity_link(time_dimension_spec: TimeDimensionSpec) -> None:  # noqa: D
    assert time_dimension_spec.without_first_entity_link == TimeDimensionSpec(
        element_name="signup_ts",
        entity_links=(),
        time_granularity=TimeGranularity.DAY,
    )


def test_time_dimension_without_entity_links(time_dimension_spec: TimeDimensionSpec) -> None:  # noqa: D
    assert time_dimension_spec.without_entity_links == TimeDimensionSpec(
        element_name="signup_ts",
        entity_links=(),
        time_granularity=time_dimension_spec.time_granularity,
    )


def test_entity_without_first_entity_link(entity_spec: EntitySpec) -> None:  # noqa: D
    assert entity_spec.without_first_entity_link == EntitySpec(
        element_name="user_id",
        entity_links=(),
    )


def test_entity_without_entity_links(entity_spec: EntitySpec) -> None:  # noqa: D
    assert entity_spec.without_entity_links == EntitySpec(
        element_name="user_id",
        entity_links=(),
    )


def test_merge_linkable_specs(dimension_spec: DimensionSpec, entity_spec: EntitySpec) -> None:  # noqa: D
    linkable_specs: Sequence[LinkableInstanceSpec] = [dimension_spec, entity_spec]

    assert LinkableInstanceSpec.merge_linkable_specs([dimension_spec], [entity_spec]) == linkable_specs


def test_qualified_name() -> None:  # noqa: D
    assert (
        DimensionSpec(element_name="country", entity_links=(EntityReference("listing_id"),)).qualified_name
        == "listing_id__country"
    )


def test_merge_spec_set() -> None:  # noqa: D
    spec_set1 = InstanceSpecSet(metric_specs=(MetricSpec(element_name="bookings"),))
    spec_set2 = InstanceSpecSet(
        dimension_specs=(DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),)
    )

    assert InstanceSpecSet.merge((spec_set1, spec_set2)) == InstanceSpecSet(
        metric_specs=(MetricSpec(element_name="bookings"),),
        dimension_specs=(DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),),
    )


@pytest.fixture
def spec_set() -> InstanceSpecSet:  # noqa: D
    return InstanceSpecSet(
        metric_specs=(MetricSpec(element_name="bookings"),),
        measure_specs=(
            MeasureSpec(
                element_name="bookings",
            ),
        ),
        dimension_specs=(DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),),
        time_dimension_specs=(
            TimeDimensionSpec(
                element_name="ds",
                entity_links=(),
                time_granularity=TimeGranularity.DAY,
            ),
        ),
        entity_specs=(
            EntitySpec(
                element_name="user_id",
                entity_links=(EntityReference(element_name="listing_id"),),
            ),
        ),
    )


def test_spec_set_linkable_specs(spec_set: InstanceSpecSet) -> None:  # noqa: D
    assert set(spec_set.linkable_specs) == {
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
        TimeDimensionSpec(
            element_name="ds",
            entity_links=(),
            time_granularity=TimeGranularity.DAY,
        ),
        EntitySpec(
            element_name="user_id",
            entity_links=(EntityReference(element_name="listing_id"),),
        ),
    }


def test_spec_set_all_specs(spec_set: InstanceSpecSet) -> None:  # noqa: D
    assert set(spec_set.all_specs) == {
        MetricSpec(element_name="bookings"),
        MeasureSpec(
            element_name="bookings",
        ),
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
        TimeDimensionSpec(
            element_name="ds",
            entity_links=(),
            time_granularity=TimeGranularity.DAY,
        ),
        EntitySpec(
            element_name="user_id",
            entity_links=(EntityReference(element_name="listing_id"),),
        ),
    }


def test_linkless_entity() -> None:  # noqa: D
    """Check that equals and hash works as expected for the LinklessEntitySpec / EntitySpec."""
    entity_spec = EntitySpec(element_name="user_id", entity_links=())
    linkless_entity_spec = LinklessEntitySpec.from_element_name("user_id")

    # Check equality between the two.
    assert entity_spec == entity_spec
    assert linkless_entity_spec == linkless_entity_spec
    assert entity_spec == linkless_entity_spec

    # Check that they are treated equivalently in sets.
    set_with_entity_spec = {entity_spec}
    assert entity_spec in set_with_entity_spec
    assert linkless_entity_spec in set_with_entity_spec

    set_with_linkless_entity_spec = {linkless_entity_spec}
    assert entity_spec in set_with_linkless_entity_spec
    assert linkless_entity_spec in set_with_linkless_entity_spec

    set_with_entity_spec.add(linkless_entity_spec)
    assert len(set_with_entity_spec) == 1
